# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Optional, Tuple

import torch
from torch import Tensor

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils.constants import Backend


class ScaledDotProductAttentionTRT(torch.autograd.Function):
    """Caller of scale dot product attention."""

    @staticmethod
    def forward(
        ctx, q: Tensor, k: Tensor, v: Tensor, attn_mask: Optional[Tensor] = None
    ):
        """forward function."""
        B, Nt, E = q.shape
        q = q / math.sqrt(E)
        # (B, Nt, E) x (B, E, Ns) -> (B, Nt, Ns)
        attn = torch.bmm(q, k.transpose(-2, -1))
        if attn_mask is not None:
            attn += attn_mask

        attn = attn.softmax(-1)

        # (B, Nt, Ns) x (B, Ns, E) -> (B, Nt, E)
        output = torch.bmm(attn, v)
        return output, attn

    @staticmethod
    def symbolic(g, q, k, v, mask):
        """Symbolic function."""
        inputs = [q, k, v]
        if mask is not None:
            inputs += [mask]
        return g.op("mmdeploy::ScaledDotProductAttentionTRT", *inputs, outputs=2)


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional._scaled_dot_product_attention",
    backend=Backend.TENSORRT.value,
)
def _scaled_dot_product_attention__tensorrt(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    attn_mask: Optional[Tensor] = None,
    dropout_p: float = 0.0,
) -> Tuple[Tensor, Tensor]:
    """Rewrite for custom ops."""
    return ScaledDotProductAttentionTRT.apply(q, k, v, attn_mask)



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.task_modules.coders.delta_xywh_bbox_coder."
    "DeltaXYWHBBoxCoder.decode",
    backend="default",
)
def deltaxywhbboxcoder__decode(
    self, bboxes, pred_bboxes, max_shape=None, wh_ratio_clip=16 / 1000
):
    """Rewrite `decode` of `DeltaXYWHBBoxCoder` for default backend.

    Rewrite this func to call `delta2bbox` directly.

    Args:
        bboxes (torch.Tensor): Basic boxes. Shape (B, N, 4) or (N, 4)
        pred_bboxes (Tensor): Encoded offsets with respect to each roi.
           Has shape (B, N, num_classes * 4) or (B, N, 4) or
           (N, num_classes * 4) or (N, 4). Note N = num_anchors * W * H
           when rois is a grid of anchors.Offset encoding follows [1]_.
        max_shape (Sequence[int] or torch.Tensor or Sequence[
           Sequence[int]],optional): Maximum bounds for boxes, specifies
           (H, W, C) or (H, W). If bboxes shape is (B, N, 4), then
           the max_shape should be a Sequence[Sequence[int]]
           and the length of max_shape should also be B.
        wh_ratio_clip (float, optional): The allowed ratio between
            width and height.

    Returns:
        torch.Tensor: Decoded boxes.
    """
    assert pred_bboxes.size(0) == bboxes.size(0)
    if pred_bboxes.ndim == 3:
        assert pred_bboxes.size(1) == bboxes.size(1)
    from mmdet.models.task_modules.coders.delta_xywh_bbox_coder import delta2bbox

    decoded_bboxes = delta2bbox(
        bboxes,
        pred_bboxes,
        self.means,
        self.stds,
        max_shape,
        wh_ratio_clip,
        self.clip_border,
        self.add_ctr_clamp,
        self.ctr_clamp,
    )
    return decoded_bboxes


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.task_modules.coders" ".delta_xywh_bbox_coder.delta2bbox",
    backend="default",
)
def delta2bbox(
    rois,
    deltas,
    means=(0.0, 0.0, 0.0, 0.0),
    stds=(1.0, 1.0, 1.0, 1.0),
    max_shape=None,
    wh_ratio_clip=16 / 1000,
    clip_border=True,
    add_ctr_clamp=False,
    ctr_clamp=32,
):
    """Rewrite `delta2bbox` for default backend.

    Since the need of clip op with dynamic min and max, this function uses
    clip_bboxes function to support dynamic shape.

    Args:
        ctx (ContextCaller): The context with additional information.
        rois (Tensor): Boxes to be transformed. Has shape (N, 4).
        deltas (Tensor): Encoded offsets relative to each roi.
            Has shape (N, num_classes * 4) or (N, 4). Note
            N = num_base_anchors * W * H, when rois is a grid of
            anchors. Offset encoding follows [1]_.
        means (Sequence[float]): Denormalizing means for delta coordinates.
            Default (0., 0., 0., 0.).
        stds (Sequence[float]): Denormalizing standard deviation for delta
            coordinates. Default (1., 1., 1., 1.).
        max_shape (tuple[int, int]): Maximum bounds for boxes, specifies
           (H, W). Default None.
        wh_ratio_clip (float): Maximum aspect ratio for boxes. Default
            16 / 1000.
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Default True.
        add_ctr_clamp (bool): Whether to add center clamp. When set to True,
            the center of the prediction bounding box will be clamped to
            avoid being too far away from the center of the anchor.
            Only used by YOLOF. Default False.
        ctr_clamp (int): the maximum pixel shift to clamp. Only used by YOLOF.
            Default 32.

    Return:
        bboxes (Tensor): Boxes with shape (N, num_classes * 4) or (N, 4),
            where 4 represent tl_x, tl_y, br_x, br_y.
    """
    means = deltas.new_tensor(means).view(1, -1)
    stds = deltas.new_tensor(stds).view(1, -1)
    delta_shape = deltas.shape
    reshaped_deltas = deltas.view(delta_shape[:-1] + (-1, 4))
    denorm_deltas = reshaped_deltas * stds + means

    dxy = denorm_deltas[..., :2]
    dwh = denorm_deltas[..., 2:]

    # fix openvino on torch1.13
    xy1 = rois[..., :2].unsqueeze(2)
    xy2 = rois[..., 2:].unsqueeze(2)

    pxy = (xy1 + xy2) * 0.5
    pwh = xy2 - xy1
    dxy_wh = pwh * dxy

    max_ratio = np.abs(np.log(wh_ratio_clip))
    if add_ctr_clamp:
        dxy_wh = torch.clamp(dxy_wh, max=ctr_clamp, min=-ctr_clamp)
        dwh = torch.clamp(dwh, max=max_ratio)
    else:
        dwh = dwh.clamp(min=-max_ratio, max=max_ratio)

    # Use exp(network energy) to enlarge/shrink each roi
    half_gwh = pwh * dwh.exp() * 0.5
    # Use network energy to shift the center of each roi
    gxy = pxy + dxy_wh

    # Convert center-xy/width/height to top-left, bottom-right
    xy1 = gxy - half_gwh
    xy2 = gxy + half_gwh

    x1 = xy1[..., 0]
    y1 = xy1[..., 1]
    x2 = xy2[..., 0]
    y2 = xy2[..., 1]

    if clip_border and max_shape is not None:
        from juxtapose.mmdeploy.codebase.mmdet.deploy import clip_bboxes

        x1, y1, x2, y2 = clip_bboxes(x1, y1, x2, y2, max_shape)

    bboxes = torch.stack([x1, y1, x2, y2], dim=-1).view(deltas.size())
    return bboxes


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.task_modules.coders." "delta_xywh_bbox_coder.delta2bbox",
    backend="ncnn",
)
def delta2bbox__ncnn(
    rois,
    deltas,
    means=(0.0, 0.0, 0.0, 0.0),
    stds=(1.0, 1.0, 1.0, 1.0),
    max_shape=None,
    wh_ratio_clip=16 / 1000,
    clip_border=True,
    add_ctr_clamp=False,
    ctr_clamp=32,
):
    """Rewrite `delta2bbox` for ncnn backend.
    Batch dimension is not supported by ncnn, but supported by pytorch.
    ncnn regards the lowest two dimensions as continuous address with byte
    alignment, so the lowest two dimensions are not absolutely independent.
    Reshape operator with -1 arguments should operates ncnn::Mat with
    dimension >= 3.
    Args:
        ctx (ContextCaller): The context with additional information.
        rois (Tensor): Boxes to be transformed. Has shape (N, 4) or (B, N, 4)
        deltas (Tensor): Encoded offsets with respect to each roi.
            Has shape (B, N, num_classes * 4) or (B, N, 4) or
            (N, num_classes * 4) or (N, 4). Note N = num_anchors * W * H
            when rois is a grid of anchors.Offset encoding follows [1]_.
        means (Sequence[float]): Denormalizing means for delta coordinates
        stds (Sequence[float]): Denormalizing standard deviation for delta
            coordinates
        max_shape (Sequence[int] or torch.Tensor or Sequence[
            Sequence[int]],optional): Maximum bounds for boxes, specifies
            (H, W, C) or (H, W). If rois shape is (B, N, 4), then
            the max_shape should be a Sequence[Sequence[int]]
            and the length of max_shape should also be B.
        wh_ratio_clip (float): Maximum aspect ratio for boxes.
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Defaults to True.
        add_ctr_clamp (bool): Whether to add center clamp, when added, the
            predicted box is clamped is its center is too far away from
            the original anchor's center. Only used by YOLOF. Default False.
        ctr_clamp (int): the maximum pixel shift to clamp. Only used by YOLOF.
            Default 32.
    Return:
        bboxes (Tensor): Boxes with shape (B, N, num_classes * 4) or (B, N, 4)
            or (N, num_classes * 4) or (N, 4), where 4 represent tl_x, tl_y,
            br_x, br_y.
    """
    means = deltas.new_tensor(means).view(1, 1, 1, -1).data
    stds = deltas.new_tensor(stds).view(1, 1, 1, -1).data
    delta_shape = deltas.shape
    reshaped_deltas = deltas.view(delta_shape[:-1] + (-1, 4))
    denorm_deltas = reshaped_deltas * stds + means

    dxy = denorm_deltas[..., :2]
    dwh = denorm_deltas[..., 2:]

    xy1 = rois[..., None, :2]
    xy2 = rois[..., None, 2:]

    pxy = (xy1 + xy2) * 0.5
    pwh = xy2 - xy1
    dxy_wh = pwh * dxy

    max_ratio = np.abs(np.log(wh_ratio_clip))
    if add_ctr_clamp:
        dxy_wh = torch.clamp(dxy_wh, max=ctr_clamp, min=-ctr_clamp)
        dwh = torch.clamp(dwh, max=max_ratio)
    else:
        dwh = dwh.clamp(min=-max_ratio, max=max_ratio)

    # Use exp(network energy) to enlarge/shrink each roi
    half_gwh = pwh * dwh.exp() * 0.5
    # Use network energy to shift the center of each roi
    gxy = pxy + dxy_wh

    # Convert center-xy/width/height to top-left, bottom-right
    xy1 = gxy - half_gwh
    xy2 = gxy + half_gwh

    x1 = xy1[..., 0]
    y1 = xy1[..., 1]
    x2 = xy2[..., 0]
    y2 = xy2[..., 1]

    if clip_border and max_shape is not None:
        from juxtapose.mmdeploy.codebase.mmdet.deploy import clip_bboxes

        x1, y1, x2, y2 = clip_bboxes(x1, y1, x2, y2, max_shape)

    bboxes = torch.stack([x1, y1, x2, y2], dim=-1).view(deltas.size())
    return bboxes



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

from juxtapose.mmdeploy.utils import Codebase
from .base import BaseTask, MMCodebase, get_codebase_class

extra_dependent_library = {Codebase.MMOCR: ["mmdet"], Codebase.MMROTATE: ["mmdet"]}


def import_codebase(codebase_type: Codebase, custom_module_list: List = []):
    """Import a codebase package in `mmdeploy.codebase`

    The function will check if all dependent libraries are installed.
    For example, to import `mmdeploy.codebase.mmdet`, `mmdet` must be
    installed. To import `mmdeploy.codebase.mmocr`, `mmdet` and `mmocr`
    must be installed.

    Args:
        codebase (Codebase): The codebase to import.
    """
    import importlib

    codebase_name = codebase_type.value

    dependent_library = [codebase_name] + extra_dependent_library.get(codebase_type, [])
    for lib in dependent_library + custom_module_list:
        if not importlib.util.find_spec(lib):
            raise ImportError(f"{lib} has not been installed. " f"Import {lib} failed.")
    if len(custom_module_list) > 0:
        for custom_module in custom_module_list:
            importlib.import_module(f"{custom_module}")
    codebase = get_codebase_class(codebase_type)
    codebase.register_all_modules()


__all__ = ["MMCodebase", "BaseTask", "get_codebase_class"]



# Copyright (c) OpenMMLab. All rights reserved.
from .extractor import create_extractor, parse_extractor_io_string
from .function_marker import mark, reset_mark_function_count
from .optimize import (attribute_to_dict, get_new_name, remove_identity,
                       remove_imports, rename_value)

__all__ = [
    'mark', 'reset_mark_function_count', 'create_extractor',
    'parse_extractor_io_string', 'remove_identity', 'attribute_to_dict',
    'rename_value', 'get_new_name', 'remove_imports'
]



# Copyright (c) OpenMMLab. All rights reserved.
import importlib
import inspect
import logging
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

from juxtapose.mmdeploy.utils import get_root_logger

try:
    import torch.multiprocessing as mp
except Exception:
    import multiprocessing as mp


def _get_func_name(func: Callable) -> str:
    """get function name."""
    assert isinstance(func, Callable), f"{func} is not a Callable object."
    _func_name = None
    if hasattr(func, "__qualname__"):
        _func_name = f"{func.__module__}.{func.__qualname__}"
    elif hasattr(func, "__class__"):
        _func_name = func.__class__
    else:
        _func_name = str(func)
    return _func_name


class PipelineCaller:
    """Classes to record the attribute of each pipeline function."""

    def __init__(
        self,
        module_name: str,
        impl_name: str,
        func_name: Optional[str] = None,
        log_level: int = logging.DEBUG,
        is_multiprocess_available: bool = True,
    ) -> None:
        if func_name is not None:
            self._func_name = func_name
        else:
            self._func_name = impl_name
        # Can not save the function directly since multiprocess with spawn mode
        # require all field can be pickled.
        self._module_name = module_name
        self._impl_name = impl_name
        self._is_multiprocess_available = is_multiprocess_available
        self._enable_multiprocess = False
        self._mp_dict = None
        self._mp_async = False
        self._call_id = 0
        self._log_level = log_level
        self._input_hooks: List[Callable] = []
        self._output_hooks: List[Callable] = []

    @property
    def is_multiprocess_available(self) -> bool:
        """check if multiprocess is available for this pipeline."""
        return self._is_multiprocess_available

    @property
    def is_multiprocess(self) -> bool:
        """check if this pipeline is multiprocess."""
        return self._enable_multiprocess

    @property
    def input_hooks(self) -> List[Callable]:
        """get input hooks."""
        return self._input_hooks

    @property
    def output_hooks(self) -> List[Callable]:
        """get output hooks."""
        return self._output_hooks

    def pop_mp_output(self, call_id: int = None) -> Any:
        """pop multiprocess output."""
        assert self._mp_dict is not None, "mp_dict is None."
        call_id = self._call_id if call_id is None else call_id
        if call_id not in self._mp_dict:
            get_root_logger().error(
                f"`{self._func_name}` with Call id: {call_id} failed. exit."
            )
            exit(1)
        ret = self._mp_dict[call_id]
        self._mp_dict.pop(call_id)
        return ret

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        do_multiprocess = (
            self.is_multiprocess_available
            and self.is_multiprocess
            and self._mp_dict is not None
        )

        logger = get_root_logger(log_level=self._log_level)
        mp_log_str = "subprocess" if do_multiprocess else "main process"
        logger.log(self._log_level, f"Start pipeline {self._func_name} in {mp_log_str}")

        for input_hook in self.input_hooks:
            args, kwargs = input_hook(*args, **kwargs)

        module_name = self._module_name
        impl_name = self._impl_name
        # TODO: find another way to load function
        mod = importlib.import_module(module_name)
        func = getattr(mod, impl_name, None)
        assert func is not None, f"Can not find implementation of {self._func_name}"
        ret = func(*args, **kwargs)
        for output_hook in self.output_hooks:
            ret = output_hook(ret)

        if do_multiprocess:
            self._mp_dict[self._call_id] = ret

        logger.log(self._log_level, f"Finish pipeline {self._func_name}")
        return ret


class PipelineResult:
    """The result of async pipeline."""

    def __init__(self, manager: Any, call_id: int) -> None:
        self._manager = manager
        self._call_id = call_id

    @property
    def call_id(self) -> int:
        return self._call_id

    def get(self) -> Any:
        """get result."""
        return self._manager.get_result_sync(self._call_id)


FUNC_NAME_TYPE = Union[str, Callable]


class PipelineManager:
    """This is a tool to manager all pipeline functions."""

    def __init__(self) -> None:
        self._enable_multiprocess = True
        self._mp_manager = None
        self._callers: Dict[str, PipelineCaller] = dict()
        self._call_id = 0
        self._proc_async: Dict[int, (str, mp.Process)] = dict()

    @property
    def mp_manager(self) -> Optional[mp.Manager]:
        """get multiprocess manager."""
        return self._mp_manager

    def get_caller(self, func_name: FUNC_NAME_TYPE) -> PipelineCaller:
        """get caller of given function."""
        if isinstance(func_name, Callable):
            func_name = _get_func_name(func_name)
        assert func_name in self._callers, f"{func_name} has not been registered."
        return self._callers[func_name]

    def __set_caller_val(
        self, val_name: str, val: Any, func_name: Optional[FUNC_NAME_TYPE] = None
    ) -> None:
        """helper to set any caller value."""
        if func_name is None:
            for func_name_ in self._callers:
                setattr(self.get_caller(func_name_), val_name, val)
        else:
            setattr(self.get_caller(func_name), val_name, val)

    def _create_mp_manager(self) -> None:
        """create multiprocess manager if not exists."""
        if self._mp_manager is None:
            self._mp_manager = mp.Manager()

    def _enable_multiprocess_single(
        self, val: bool, func_name: FUNC_NAME_TYPE = None
    ) -> None:
        """implement of enable_multiprocess."""
        pipe_caller = self.get_caller(func_name)
        # check if multiprocess is available for this function
        if not pipe_caller.is_multiprocess_available:
            return
        pipe_caller._enable_multiprocess = val
        if val is True and self.mp_manager is not None:
            pipe_caller._mp_dict = self.mp_manager.dict()
        else:
            pipe_caller._mp_dict = None

    def enable_multiprocess(
        self,
        val: bool,
        func_names: Optional[Union[FUNC_NAME_TYPE, Sequence[FUNC_NAME_TYPE]]] = None,
    ) -> None:
        """enable multiprocess for pipeline function.

        Args:
            val (bool): enable or disable multiprocess.
            func_names (str | List[str]): function names to enable. If
                func_name is None, all registered function will be enabled.
        """
        if val is True:
            self._create_mp_manager()
        if func_names is None:
            for func_name in self._callers:
                self._enable_multiprocess_single(val, func_name=func_name)
        else:
            if isinstance(func_names, str):
                func_names = [func_names]
            for func_name in func_names:
                self._enable_multiprocess_single(val, func_name=func_name)

    def set_mp_async(
        self, val: bool, func_name: Optional[FUNC_NAME_TYPE] = None
    ) -> None:
        """set multiprocess async of the pipeline function.

        Args:
            val (bool): enable async call.
            func_name (str | None): function name to set. If func_name is
                None, all registered function will be set.
        """
        self.__set_caller_val("_mp_async", val, func_name)

    def set_log_level(
        self,
        level: int,
        func_names: Optional[Union[FUNC_NAME_TYPE, Sequence[FUNC_NAME_TYPE]]] = None,
    ) -> None:
        """set log level of the pipeline function.

        Args:
            level (int): the log level.
            func_names (str | List[str]): function names to set. If func_names
                is None, all registered function will be set.
        """
        if isinstance(func_names, str):
            func_names = [func_names]
        for func_name in func_names:
            self.__set_caller_val("_log_level", level, func_name)

    def get_input_hooks(self, func_name: FUNC_NAME_TYPE):
        """get input hooks of given function name.

        Args:
            func_name (str): function name.
        """
        pipe_caller = self.get_caller(func_name)
        return pipe_caller.input_hooks

    def get_output_hooks(self, func_name: FUNC_NAME_TYPE):
        """get output hooks of given function name.

        Args:
            func_name (str): function name.
        """
        pipe_caller = self.get_caller(func_name)
        return pipe_caller.output_hooks

    def call_function_local(self, func_name: FUNC_NAME_TYPE, *args, **kwargs) -> Any:
        """call pipeline function.

        Args:
            func_name (str): function name to be called.

        Returns:
            Any: The result of call function
        """
        pipe_caller = self.get_caller(func_name)
        pipe_caller._call_id = self._call_id
        self._call_id += 1
        return pipe_caller(*args, **kwargs)

    def call_function_async(self, func_name: FUNC_NAME_TYPE, *args, **kwargs) -> int:
        """call pipeline function.

        Args:
            func_name (str): function name to be called.

        Returns:
            int: Call id of this function
        """
        pipe_caller = self.get_caller(func_name)
        assert (
            pipe_caller.is_multiprocess
        ), f"multiprocess of {func_name} has not been enabled."

        call_id = self._call_id
        pipe_caller._call_id = call_id
        self._call_id += 1
        proc = mp.Process(target=pipe_caller, args=args, kwargs=kwargs)
        proc.start()
        self._proc_async[call_id] = (func_name, proc)

        return call_id

    def get_result_sync(self, call_id: int):
        """get result of async call."""
        assert call_id in self._proc_async, f"Unknown call id: {call_id}"
        func_name, proc = self._proc_async.pop(call_id)
        proc.join()
        ret = self.get_caller(func_name).pop_mp_output(call_id)

        return ret

    def call_function(self, func_name: FUNC_NAME_TYPE, *args, **kwargs) -> Any:
        """call pipeline function.

        Args:
            func_name (str): function name to be called.

        Returns:
            Any: The result of call function
        """
        pipe_caller = self.get_caller(func_name)

        if self._enable_multiprocess and pipe_caller.is_multiprocess:
            call_id = self.call_function_async(func_name, *args, **kwargs)
            if pipe_caller._mp_async:
                return PipelineResult(self, call_id)
            return self.get_result_sync(call_id)
        else:
            return self.call_function_local(func_name, *args, **kwargs)

    def register_pipeline(
        self, is_multiprocess_available: bool = True, log_level: int = logging.DEBUG
    ):
        """register the pipeline function."""

        def _register(func):
            assert isinstance(func, Callable), f"{func} is not Callable."
            func_name_ = _get_func_name(func)

            # save the implementation into the registry module
            impl_name = f"_pipe_{func.__name__}__impl_"
            frame = inspect.stack()[1]
            outer_mod = inspect.getmodule(frame[0])
            mod_name = outer_mod.__name__
            setattr(outer_mod, impl_name, func)

            # create caller
            pipe_caller = PipelineCaller(
                mod_name,
                impl_name,
                func_name=func_name_,
                log_level=log_level,
                is_multiprocess_available=is_multiprocess_available,
            )
            PIPELINE_MANAGER._callers[func_name_] = pipe_caller

            # wrap call
            @wraps(func)
            def _wrap(*args, **kwargs):
                return self.call_function(func_name_, *args, **kwargs)

            return _wrap

        return _register


PIPELINE_MANAGER = PipelineManager()


class no_mp:
    """The context manager used to disable multiprocess."""

    def __init__(self, manager: PipelineManager = PIPELINE_MANAGER) -> None:
        self._manager = manager
        self._old_enable_multiprocess = True

    def __enter__(self):
        self._old_enable_multiprocess = self._manager._enable_multiprocess
        self._manager._enable_multiprocess = False

    def __exit__(self, type, val, tb):
        self._manager._enable_multiprocess = self._old_enable_multiprocess



# Copyright (c) OpenMMLab. All rights reserved.
import re
from typing import Optional


def parse_device_id(device: str) -> Optional[int]:
    """Parse device index from a string.

    Args:
        device (str): The typical style of string specifying device,
            e.g.: 'cuda:0', 'cpu'.

    Returns:
        Optional[int]: The return value depends on the type of device.
            If device is 'cuda': cuda device index, defaults to `0`.
            If device is 'cpu': `-1`.
            Otherwise, `None` will be returned.
    """
    if device == 'cpu':
        return -1
    if 'cuda' in device:
        return parse_cuda_device_id(device)
    return None


def parse_cuda_device_id(device: str) -> int:
    """Parse cuda device index from a string.

    Args:
        device (str): The typical style of string specifying cuda device,
            e.g.: 'cuda:0'.

    Returns:
        int: The parsed device id, defaults to `0`.
    """
    match_result = re.match('([^:]+)(:[0-9]+)?$', device)
    assert match_result is not None, f'Can not parse device {device}.'
    assert match_result.group(1).lower() == 'cuda', 'Not cuda device.'

    device_id = 0 if match_result.lastindex == 1 else int(
        match_result.group(2)[1:])

    return device_id


def parse_device_type(device: str) -> str:
    """Parse device type from a string.

    Args:
        device (str): The typical style of string specifying cuda device,
            e.g.: 'cuda:0', 'cpu', 'npu'.

    Returns:
        str: The parsed device type such as 'cuda', 'cpu', 'npu'.
    """
    device_type = device
    if ':' in device:
        device_type = device.split(':')[0]
    return device_type



# Copyright (c) OpenMMLab. All rights reserved.
import importlib

from juxtapose.mmdeploy.utils import get_root_logger
from .version import __version__, version_info  # noqa F401

if importlib.util.find_spec("torch"):
    importlib.import_module("juxtapose.mmdeploy.pytorch")
else:
    logger = get_root_logger()
    logger.debug("torch is not installed.")

if importlib.util.find_spec("mmcv"):
    importlib.import_module("juxtapose.mmdeploy.mmcv")
else:
    logger = get_root_logger()
    logger.debug("mmcv is not installed.")



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn.functional as F

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmcv.ops.point_sample", backend="default"
)
def point_sample__default(input, points, align_corners=False, **kwargs):
    """A wrapper around :func:`grid_sample` to support 3D point_coords tensors
    Unlike :func:`torch.nn.functional.grid_sample` it assumes point_coords to
    lie inside ``[0, 1] x [0, 1]`` square.

    Args:
        input (torch.Tensor): Feature map, shape (N, C, H, W).
        points (torch.Tensor): Image based absolute point coordinates
            (normalized), range [0, 1] x [0, 1], shape (N, P, 2) or
            (N, Hgrid, Wgrid, 2).
        align_corners (bool, optional): Whether align_corners.
            Default: False

    Returns:
        torch.Tensor: Features of `point` on `input`, shape (N, C, P) or
        (N, C, Hgrid, Wgrid).
    """
    from mmcv.ops.point_sample import denormalize

    add_dim = False
    if points.dim() == 3:
        add_dim = True
        points = points.unsqueeze(2)
    output = F.grid_sample(
        input, denormalize(points), align_corners=align_corners, **kwargs
    )
    if add_dim:
        output = output.squeeze(3)
    return output


@FUNCTION_REWRITER.register_rewriter(func_name="mmcv.ops.SimpleRoIAlign.forward")
def simple_roialign__forward(self, features, rois):
    """Rewrite `forward` of SimpleRoIAlign.

    Args:
        features (torch.Tensor): Feature map, shape (N, C, H, W).
        rois (torch.Tensor):

    Returns:
        torch.Tensor: RoI features.
    """
    from mmcv.ops.point_sample import (
        generate_grid,
        point_sample,
        rel_roi_point_to_rel_img_point,
    )

    num_imgs = features.size(0)
    num_rois = rois.size(0)
    rel_roi_points = generate_grid(num_rois, self.output_size, device=rois.device)
    rel_img_points = rel_roi_point_to_rel_img_point(
        rois, rel_roi_points, features, self.spatial_scale
    )
    rel_img_points = rel_img_points.reshape(num_imgs, -1, *rel_img_points.shape[1:])
    point_feats = point_sample(features, rel_img_points, align_corners=not self.aligned)
    point_feats = point_feats.transpose(1, 2)

    channels = features.size(1)
    roi_feats = point_feats.reshape(num_rois, channels, *self.output_size)

    return roi_feats



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Tuple

import torch
from mmdet.utils import ConfigType
from torch import Tensor

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.standard_roi_head.StandardRoIHead.predict_bbox"
)
def standard_roi_head__predict_bbox(
    self,
    x: Tuple[Tensor],
    batch_img_metas: List[dict],
    rpn_results_list: List[Tensor],
    rcnn_test_cfg: ConfigType,
    rescale: bool = False,
) -> List[Tensor]:
    """Rewrite `predict_bbox` of `StandardRoIHead` for default backend.

    Args:
        x (tuple[Tensor]): Feature maps of all scale level.
        batch_img_metas (list[dict]): List of image information.
        rpn_results_list (list[Tensor]): List of region
            proposals.
        rcnn_test_cfg (obj:`ConfigDict`): `test_cfg` of R-CNN.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.

    Returns:
        list[Tensor]: Detection results of each image
        after the post process.
        Each item usually contains following keys.

            - dets (Tensor): Classification bboxes and scores, has a shape
                (num_instance, 5)
            - labels (Tensor): Labels of bboxes, has a shape
                (num_instances, ).
    """
    rois = rpn_results_list[0]
    rois_dims = int(rois.shape[-1])
    batch_index = (
        torch.arange(rois.shape[0], device=rois.device)
        .float()
        .view(-1, 1, 1)
        .expand(rois.size(0), rois.size(1), 1)
    )
    rois = torch.cat([batch_index, rois[..., : rois_dims - 1]], dim=-1)
    batch_size = rois.shape[0]
    num_proposals_per_img = rois.shape[1]

    # Eliminate the batch dimension
    rois = rois.view(-1, rois_dims)
    bbox_results = self._bbox_forward(x, rois)
    cls_scores = bbox_results["cls_score"]
    bbox_preds = bbox_results["bbox_pred"]

    # Recover the batch dimension
    rois = rois.reshape(batch_size, num_proposals_per_img, rois.size(-1))
    cls_scores = cls_scores.reshape(
        batch_size, num_proposals_per_img, cls_scores.size(-1)
    )

    bbox_preds = bbox_preds.reshape(
        batch_size, num_proposals_per_img, bbox_preds.size(-1)
    )
    result_list = self.bbox_head.predict_by_feat(
        rois=rois,
        cls_scores=cls_scores,
        bbox_preds=bbox_preds,
        batch_img_metas=batch_img_metas,
        rcnn_test_cfg=rcnn_test_cfg,
        rescale=rescale,
    )
    return result_list


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.roi_heads.standard_roi_head.StandardRoIHead.predict_mask"
)
def standard_roi_head__predict_mask(
    self,
    x: Tuple[Tensor],
    batch_img_metas: List[dict],
    results_list: List[Tensor],
    rescale: bool = False,
) -> List[Tensor]:
    """Perform forward propagation of the mask head and predict detection
    results on the features of the upstream network.

    Args:
        x (tuple[Tensor]): Feature maps of all scale level.
        batch_img_metas (list[dict]): List of image information.
        results_list (list[:obj:`InstanceData`]): Detection results of
            each image.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.

    Returns:
        list[Tensor]: Detection results of each image
        after the post process.
        Each item usually contains following keys.

            - scores (Tensor): Classification scores, has a shape
                (num_instance, )
            - labels (Tensor): Labels of bboxes, has a shape
                (num_instances, ).
            - bboxes (Tensor): Has a shape (num_instances, 4),
                the last dimension 4 arrange as (x1, y1, x2, y2).
            - masks (Tensor): Has a shape (num_instances, H, W).
    """
    dets, det_labels = results_list
    batch_size = dets.size(0)
    det_bboxes = dets[..., :4]
    # expand might lead to static shape, use broadcast instead
    batch_index = torch.arange(
        det_bboxes.size(0), device=det_bboxes.device
    ).float().view(-1, 1, 1) + det_bboxes.new_zeros(
        (det_bboxes.size(0), det_bboxes.size(1))
    ).unsqueeze(
        -1
    )
    mask_rois = torch.cat([batch_index, det_bboxes], dim=-1)
    mask_rois = mask_rois.view(-1, 5)
    mask_results = self._mask_forward(x, mask_rois)
    mask_preds = mask_results["mask_preds"]
    num_det = det_bboxes.shape[1]
    segm_results = self.mask_head.predict_by_feat(
        mask_preds, results_list, batch_img_metas, self.test_cfg, rescale=rescale
    )
    segm_results = segm_results.reshape(
        batch_size, num_det, segm_results.shape[-2], segm_results.shape[-1]
    )
    return dets, det_labels, segm_results



backend_config = dict(
    type='onnxruntime',
    precision='fp16',
    common_config=dict(
        min_positive_val=1e-7,
        max_finite_val=1e4,
        keep_io_types=False,
        disable_shape_infer=False,
        op_block_list=None,
        node_block_list=None))



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.backend.vacc import is_available
from ..core import PIPELINE_MANAGER

__all__ = ["is_available"]

if is_available():
    try:
        from juxtapose.mmdeploy.backend.vacc import from_onnx as _from_onnx

        from_onnx = PIPELINE_MANAGER.register_pipeline()(_from_onnx)
        __all__ += ["from_onnx"]
    except Exception:
        pass



# Copyright (c) OpenMMLab. All rights reserved.
import copy

import torch
from mmdet.models.detectors.base import ForwardResults
from mmdet.structures.det_data_sample import OptSampleList

from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.utils import is_dynamic_shape


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.detectors.two_stage.TwoStageDetector.extract_feat"
)
def two_stage_detector__extract_feat(self, img):
    """Rewrite `extract_feat` for default backend.

    This function uses the specific `extract_feat` function for the two
    stage detector after adding marks.

    Args:
        ctx (ContextCaller): The context with additional information.
        self: The instance of the original class.
        img (Tensor | List[Tensor]): Input image tensor(s).

    Returns:
        list[Tensor]: Each item with shape (N, C, H, W) corresponds one
        level of backbone and neck features.
    """
    ctx = FUNCTION_REWRITER.get_context()

    @mark("extract_feat", inputs="img", outputs="feat")
    def __extract_feat_impl(self, img):
        return ctx.origin_func(self, img)

    return __extract_feat_impl(self, img)


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.detectors.two_stage.TwoStageDetector.forward"
)
def two_stage_detector__forward(
    self,
    batch_inputs: torch.Tensor,
    data_samples: OptSampleList = None,
    mode: str = "tensor",
    **kwargs,
) -> ForwardResults:
    """Rewrite `forward` for default backend.

    Support configured dynamic/static shape for model input and return
    detection result as Tensor instead of numpy array.

    Args:
        batch_inputs (Tensor): Inputs with shape (N, C, H, W).
        data_samples (List[:obj:`DetDataSample`]): The Data
            Samples. It usually includes information such as
            `gt_instance`, `gt_panoptic_seg` and `gt_sem_seg`.
        mode (str): export mode, not used.

    Returns:
        tuple[Tensor]: Detection results of the
        input images.
            - dets (Tensor): Classification bboxes and scores.
                Has a shape (num_instances, 5)
            - labels (Tensor): Labels of bboxes, has a shape
                (num_instances, ).
    """
    ctx = FUNCTION_REWRITER.get_context()
    data_samples = copy.deepcopy(data_samples)
    deploy_cfg = ctx.cfg

    # get origin input shape as tensor to support onnx dynamic shape
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    img_shape = torch._shape_as_tensor(batch_inputs)[2:]
    if not is_dynamic_flag:
        img_shape = [int(val) for val in img_shape]

    # set the metainfo
    # note that we can not use `set_metainfo`, deepcopy would crash the
    # onnx trace.
    for data_sample in data_samples:
        data_sample.set_field(name="img_shape", value=img_shape, field_type="metainfo")

    x = self.extract_feat(batch_inputs)

    if data_samples[0].get("proposals", None) is None:
        rpn_results_list = self.rpn_head.predict(x, data_samples, rescale=False)
    else:
        rpn_results_list = [data_sample.proposals for data_sample in data_samples]

    output = self.roi_head.predict(x, rpn_results_list, data_samples, rescale=False)
    return output



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import Optional, Union

from mmengine import Config

from .core import PIPELINE_MANAGER, no_mp


@PIPELINE_MANAGER.register_pipeline()
def create_calib_input_data(
    calib_file: str,
    deploy_cfg: Union[str, Config],
    model_cfg: Union[str, Config],
    model_checkpoint: Optional[str] = None,
    dataset_cfg: Optional[Union[str, Config]] = None,
    dataset_type: str = "val",
    device: str = "cpu",
) -> None:
    """Create dataset for post-training quantization.

    Args:
        calib_file (str): The output calibration data file.
        deploy_cfg (str | Config): Deployment config file or
            Config object.
        model_cfg (str | Config): Model config file or Config object.
        model_checkpoint (str): A checkpoint path of PyTorch model,
            defaults to `None`.
        dataset_cfg (Optional[Union[str, Config]], optional): Model
            config to provide calibration dataset. If none, use `model_cfg`
            as the dataset config. Defaults to None.
        dataset_type (str, optional): The dataset type. Defaults to 'val'.
        device (str, optional): Device to create dataset. Defaults to 'cpu'.
    """

    from juxtapose.mmdeploy.core import patch_model
    from juxtapose.mmdeploy.utils import (
        IR,
        cfg_apply_marks,
        get_backend,
        get_ir_config,
        load_config,
    )
    from .utils import create_calib_input_data as create_calib_input_data_impl

    with no_mp():
        if dataset_cfg is None:
            dataset_cfg = model_cfg

        # load cfg if necessary
        deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)

        if dataset_cfg is None:
            dataset_cfg = model_cfg

        # load dataset_cfg if necessary
        dataset_cfg = load_config(dataset_cfg)[0]
        calib_dataloader = deepcopy(dataset_cfg[f"{dataset_type}_dataloader"])
        calib_dataloader["batch_size"] = 1

        from juxtapose.mmdeploy.apis.utils import build_task_processor

        task_processor = build_task_processor(model_cfg, deploy_cfg, device)

        apply_marks = cfg_apply_marks(deploy_cfg)

        model = task_processor.build_pytorch_model(model_checkpoint)
        dataset = task_processor.build_dataset(calib_dataloader["dataset"])
        calib_dataloader["dataset"] = dataset
        dataloader = task_processor.build_dataloader(calib_dataloader)

        # patch model
        backend = get_backend(deploy_cfg).value
        ir = IR.get(get_ir_config(deploy_cfg)["type"])
        patched_model = patch_model(model, cfg=deploy_cfg, backend=backend, ir=ir)

        def get_tensor_func(input_data):
            input_data = model.data_preprocessor(input_data)
            return input_data["inputs"]

        create_calib_input_data_impl(
            calib_file,
            patched_model,
            dataloader,
            get_tensor_func=get_tensor_func,
            inference_func=model.forward,
            model_partition=apply_marks,
            context_info=dict(cfg=deploy_cfg),
            device=device,
        )



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.backend.sdk import is_available

__all__ = ["is_available"]

if is_available():
    try:
        from juxtapose.mmdeploy.backend.sdk.export_info import export2SDK as _export2SDK
        from ..core import PIPELINE_MANAGER

        export2SDK = PIPELINE_MANAGER.register_pipeline()(_export2SDK)

        __all__ += ["export2SDK"]
    except Exception:
        pass



onnx_config = dict(
    type='onnx',
    export_params=True,
    keep_initializers_as_inputs=False,
    opset_version=11,
    save_file='end2end.onnx',
    input_names=['input'],
    output_names=['output'],
    input_shape=None,
    optimize=True)



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.backend.rknn import is_available

__all__ = ["is_available"]

if is_available():
    from juxtapose.mmdeploy.backend.rknn.onnx2rknn import onnx2rknn as _onnx2rknn
    from ..core import PIPELINE_MANAGER

    onnx2rknn = PIPELINE_MANAGER.register_pipeline()(_onnx2rknn)

    __all__ += ["onnx2rknn"]



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(func_name="torch.atan2", backend="default")
def atan2__default(
    input1: torch.Tensor,
    input2: torch.Tensor,
):
    """Rewrite `atan2` for default backend."""
    return torch.atan(input1 / (input2 + 1e-6))



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend, get_root_logger


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.necks.ssd_neck.L2Norm.forward"
)
def l2norm__forward__default(self, x):
    """Default rewriter for l2norm.

    Implement with functinoal.normalize .
    """
    return torch.nn.functional.normalize(x, dim=1) * self.weight[None, :, None, None]


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.necks.ssd_neck.L2Norm.forward",
    backend=Backend.TENSORRT.value,
)
def l2norm__forward__tensorrt(self, x):
    """rewrite `l2norm` for TensorRT.

    TensorRT7 does not support dynamic clamp, which is used in normalize.
    """
    ctx = FUNCTION_REWRITER.get_context()
    logger = get_root_logger()
    trt_version_major = 8
    try:
        import tensorrt as trt
        from packaging import version

        trt_version = version.parse(trt.__version__)
        trt_version_major = trt_version.major
    except Exception:
        logger.warning("Can not get TensorRT version.")
    if trt_version_major >= 8:
        return l2norm__forward__default(self, x)
    else:
        return ctx.origin_func(self, x)



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.clip", backend=Backend.COREML.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.clip", backend=Backend.COREML.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.clamp", backend=Backend.COREML.value
)
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.clamp", backend=Backend.COREML.value
)
def clip__coreml(input, min=None, max=None, **kwargs) -> torch.Tensor:
    """Rewrite `clip` for coreml backend.

    Cast data type.
    """
    ctx = FUNCTION_REWRITER.get_context()
    if min is not None and not isinstance(min, torch.Tensor):
        min = input.new_tensor(min)

    if max is not None and not isinstance(max, torch.Tensor):
        max = input.new_tensor(max)

    return ctx.origin_func(input, min=min, max=max, **kwargs)



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine import Registry

from juxtapose.mmdeploy.utils.config_utils import Backend


def __build_backend_wrapper_class(backend: Backend, registry: Registry):
    return registry.module_dict[backend.value]


BACKEND_WRAPPER = Registry("backend", __build_backend_wrapper_class)


def get_backend_wrapper_class(backend: Backend) -> type:
    """Get the backend wrapper class from the registry.

    Args:
        backend (Backend): The backend enum type.

    Returns:
        type: The backend wrapper class
    """
    return BACKEND_WRAPPER.build(backend)


def get_backend_file_count(backend: Backend):
    backend_class = get_backend_wrapper_class(backend)
    return backend_class.get_backend_file_count()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence, Union

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.repeat", backend="tensorrt"
)
def tensor__repeat__tensorrt(
    input: torch.Tensor, *size: Union[torch.Size, Sequence[int]]
):
    """Rewrite `repeat` for TensorRT backend.

    Some layers in TensorRT can not be applied on batch axis. add extra axis
    before operation and remove it afterward.
    """
    ctx = FUNCTION_REWRITER.get_context()

    origin_func = ctx.origin_func
    if input.dim() == 1 and len(size) == 1:
        return origin_func(input.unsqueeze(0), *([1] + list(size))).squeeze(0)
    else:
        return origin_func(input, *size)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Union

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


class GemmOp(torch.autograd.Function):
    """Create onnx::Gemm op."""

    @staticmethod
    def forward(ctx, input, weight, bias=None):
        out = input @ weight.transpose(0, 1)
        if bias is not None:
            out += bias
        return out

    @staticmethod
    def symbolic(g, input, weight, bias=None):
        input.setDebugName("A")
        weight.setDebugName("B")
        args = ["Gemm", input, weight]
        if bias is not None:
            bias.setDebugName("C")
            args.append(bias)
        return g.op(*args, alpha_f=1.0, beta_f=1.0, transA_i=0, transB_i=1)


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.linear", backend="ncnn"
)
def linear__ncnn(
    input: torch.Tensor,
    weight: torch.Tensor,
    bias: Optional[Union[torch.Tensor, torch.NoneType]] = None,
):
    """Rewrite `linear` for ncnn backend.

    The broadcast rules are different between ncnn and PyTorch. This function
    add extra reshape and transpose to support linear operation of different
    input shape.
    """
    ctx = FUNCTION_REWRITER.get_context()
    origin_func = ctx.origin_func
    dim = input.dim()

    if dim == 2 or dim == 3 and input.shape[0] == 1:
        # export nn.linear to Gemm op in onnx
        return GemmOp.apply(input, weight, bias)
    else:
        out = origin_func(input, weight)

        # permute
        out = out.transpose(1, dim - 1)

        # ncnn only support [c, h, w] and [c, 1, 1] broadcast
        out_shape = out.shape
        batch_size = out_shape[0]
        broad_cast_size = out_shape[1]
        out = out.reshape([batch_size, broad_cast_size, -1, 1])

        # add bias
        if bias is not None:
            bias = bias.view([1, -1, 1, 1])
            out = out + bias

        # permute back
        # the last dim should be -1 to support dynamic shape
        out = out.reshape(out_shape[:-1] + (-1,))
        out = out.transpose(1, dim - 1)
        return out



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

from torch import Tensor

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.dense_heads.centernet_head.CenterNetHead.predict_by_feat"
)
def centernet_head__predict_by_feat__default(
    self,
    center_heatmap_preds: List[Tensor],
    wh_preds: List[Tensor],
    offset_preds: List[Tensor],
    batch_img_metas: List[dict],
    rescale: bool = True,
    with_nms: bool = False,
):
    """Rewrite `centernethead` of `CenterNetHead` for default backend."""

    # The dynamic shape deploy of CenterNet get wrong result on TensorRT-8.4.x
    # because of TensorRT bugs, https://github.com/NVIDIA/TensorRT/issues/2299,
    # FYI.

    assert len(center_heatmap_preds) == len(wh_preds) == len(offset_preds) == 1
    batch_center_heatmap_preds = center_heatmap_preds[0]
    batch_wh_preds = wh_preds[0]
    batch_offset_preds = offset_preds[0]
    batch_size = batch_center_heatmap_preds.shape[0]
    img_shape = batch_img_metas[0]["img_shape"]
    batch_det_bboxes, batch_labels = self._decode_heatmap(
        batch_center_heatmap_preds,
        batch_wh_preds,
        batch_offset_preds,
        img_shape,
        k=self.test_cfg.topk,
        kernel=self.test_cfg.local_maximum_kernel,
    )
    det_bboxes = batch_det_bboxes.reshape([batch_size, -1, 5])
    det_labels = batch_labels.reshape(batch_size, -1)

    if with_nms:
        det_bboxes, det_labels = self._bboxes_nms(det_bboxes, det_labels, self.test_cfg)
    return det_bboxes, det_labels



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.backend.snpe import from_onnx as _from_onnx
from juxtapose.mmdeploy.backend.snpe import is_available
from ..core import PIPELINE_MANAGER

from_onnx = PIPELINE_MANAGER.register_pipeline()(_from_onnx)

__all__ = ["is_available", "from_onnx"]

if is_available():
    try:
        from juxtapose.mmdeploy.backend.snpe.onnx2dlc import (
            get_env_key,
            get_output_model_file,
        )

        __all__ += ["get_output_model_file", "get_env_key"]
    except Exception:
        pass



# Copyright (c) OpenMMLab. All rights reserved.
import time
import warnings
from contextlib import contextmanager
from logging import Logger
from typing import Optional

import numpy as np
import torch

from juxtapose.mmdeploy.utils.logging import get_logger


class TimeCounter:
    """A tool for counting inference time of backends."""

    names = dict()

    # Avoid instantiating every time
    @classmethod
    def count_time(
        cls, name: str, warmup: int = 1, log_interval: int = 1, with_sync: bool = False
    ):
        """Proceed time counting.

        Args:
            name (str): Name of this timer.
            warmup (int): The warm up steps, default 1.
            log_interval (int): Interval between each log, default 1.
            with_sync (bool): Whether use cuda synchronize for time counting,
                default `False`.
        """

        def _register(func):
            assert warmup >= 1
            assert (
                name not in cls.names
            ), "The registered function name cannot be repeated!"
            # When adding on multiple functions, we need to ensure that the
            # data does not interfere with each other
            cls.names[name] = dict(
                count=0,
                execute_time=[],
                log_interval=log_interval,
                warmup=warmup,
                with_sync=with_sync,
                batch_size=1,
                enable=False,
            )

            def fun(*args, **kwargs):
                count = cls.names[name]["count"]
                execute_time = cls.names[name]["execute_time"]
                log_interval = cls.names[name]["log_interval"]
                warmup = cls.names[name]["warmup"]
                with_sync = cls.names[name]["with_sync"]
                batch_size = cls.names[name]["batch_size"]
                enable = cls.names[name]["enable"]

                count += 1
                cls.names[name]["count"] = count

                if enable:
                    if with_sync and torch.cuda.is_available():
                        torch.cuda.synchronize()
                    start_time = time.perf_counter()

                result = func(*args, **kwargs)

                if enable:
                    if with_sync and torch.cuda.is_available():
                        torch.cuda.synchronize()
                    elapsed = (time.perf_counter() - start_time) / batch_size

                if enable and count > warmup:
                    execute_time.append(elapsed)

                    if (count - warmup) % log_interval == 0:
                        times_per_count = 1000 * float(np.mean(execute_time))
                        fps = 1000 / times_per_count
                        msg = (
                            f"[{name}]-{count} times per count: "
                            f"{times_per_count:.2f} ms, "
                            f"{fps:.2f} FPS"
                        )
                        cls.logger.info(msg)

                return result

            return fun

        return _register

    @classmethod
    @contextmanager
    def activate(
        cls,
        func_name: str = None,
        warmup: int = 1,
        log_interval: int = 1,
        with_sync: bool = False,
        file: Optional[str] = None,
        logger: Optional[Logger] = None,
        batch_size: int = 1,
        **kwargs,
    ):
        """Activate the time counter.

        Args:
            func_name (str): Specify which function to activate. If not
                specified, all registered function will be activated.
            warmup (int): the warm up steps, default 1.
            log_interval (int): Interval between each log, default 1.
            with_sync (bool): Whether use cuda synchronize for time counting,
                default False.
            file (str | None): The file to save output messages. The default
                is `None`.
            logger (Logger): The logger for the timer. Default to None.
            batch_size (int): The batch size. Default to 1.
        """
        assert warmup >= 1
        if logger is None:
            logger = get_logger("test", log_file=file)
        cls.logger = logger
        if func_name is not None:
            warnings.warn(
                "func_name must be globally unique if you call "
                "activate multiple times"
            )
            assert (
                func_name in cls.names
            ), "{} must be registered before " "setting params".format(func_name)
            cls.names[func_name]["warmup"] = warmup
            cls.names[func_name]["log_interval"] = log_interval
            cls.names[func_name]["with_sync"] = with_sync
            cls.names[func_name]["batch_size"] = batch_size
            cls.names[func_name]["enable"] = True
        else:
            for name in cls.names:
                cls.names[name]["warmup"] = warmup
                cls.names[name]["log_interval"] = log_interval
                cls.names[name]["with_sync"] = with_sync
                cls.names[name]["batch_size"] = batch_size
                cls.names[name]["enable"] = True
        yield
        if func_name is not None:
            cls.names[func_name]["enable"] = False
        else:
            for name in cls.names:
                cls.names[name]["enable"] = False

    @classmethod
    def print_stats(cls, name: str):
        """print statistics results of timer.

        Args:
            name (str): The name registered with `count_time`.
        """
        from prettytable import PrettyTable

        assert name in cls.names
        stats = cls.names[name]
        execute_time = stats["execute_time"]
        latency_mean = 1000 * float(np.mean(execute_time))
        latency_median = 1000 * float(np.median(execute_time))
        latency_min = 1000 * float(np.min(execute_time))
        latency_max = 1000 * float(np.max(execute_time))
        fps_mean, fps_median = 1000 / latency_mean, 1000 / latency_median
        fps_min, fps_max = 1000 / latency_min, 1000 / latency_max
        results = PrettyTable()
        results.field_names = ["Stats", "Latency/ms", "FPS"]
        results.add_rows(
            [
                ["Mean", latency_mean, fps_mean],
                ["Median", latency_median, fps_median],
                ["Min", latency_min, fps_min],
                ["Max", latency_max, fps_max],
            ]
        )
        results.float_format = ".3"
        print(results)



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import Backend


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmcv.cnn.bricks.transformer.PatchEmbed.forward",
    backend=Backend.NCNN.value,
)
def patch_embed__forward__ncnn(self, x):
    """Rewrite `forward` of PatchEmbed for ncnn backend.

    Args:
        x (Tensor): Has shape (B, C, H, W). In most case, C is 3.

    Returns:
        tuple: Contains merged results and its spatial shape.

        - x (Tensor): Has shape (B, out_h * out_w, embed_dims)
        - out_size (tuple[int]): Spatial shape of x, arrange as
            (out_h, out_w).
    """

    if self.adaptive_padding:
        x = self.adaptive_padding(x)

    x = self.projection(x)
    x_shape = x.shape
    out_size = (x_shape[2], x_shape[3])
    x = x.reshape((x_shape[0], x_shape[1], -1)).transpose(1, 2)
    if self.norm is not None:
        x = self.norm(x)
    return x, out_size



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER


@SYMBOLIC_REWRITER.register_symbolic("mmcv.ops.deform_conv.DeformConv2dFunction")
def deform_conv__default(
    g,
    input,
    offset,
    weight,
    stride,
    padding,
    dilation,
    groups,
    deform_groups,
    bias=False,
    im2col_step=32,
):
    """Rewrite symbolic function for default backend."""
    return g.op(
        "mmdeploy::MMCVDeformConv2d",
        input,
        offset,
        weight,
        stride_i=stride,
        padding_i=[p for pair in zip(padding, padding) for p in pair],
        dilation_i=dilation,
        groups_i=groups,
        deform_groups_i=deform_groups,
    )


@SYMBOLIC_REWRITER.register_symbolic(
    "mmcv.ops.deform_conv.DeformConv2dFunction", backend="openvino"
)
def deform_conv_openvino(
    g,
    input,
    offset,
    weight,
    stride,
    padding,
    dilation,
    groups,
    deform_groups,
    bias=False,
    im2col_step=32,
):
    """Rewrite symbolic function for OpenVINO backend."""
    assert not bias, 'The "bias" parameter should be False.'
    assert groups == 1, 'The "groups" parameter should be 1.'
    kh, kw = weight.type().sizes()[2:]
    domain = "org.openvinotoolkit"
    op_name = "DeformableConv2D"
    return g.op(
        f"{domain}::{op_name}",
        input,
        offset,
        weight,
        strides_i=stride,
        pads_i=[p for pair in zip(padding, padding) for p in pair],
        dilations_i=dilation,
        groups_i=groups,
        deformable_groups_i=deform_groups,
        kernel_shape_i=[kh, kw],
    )



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    "mmpose.models.pose_estimators.base.BasePoseEstimator.forward"
)
def base_pose_estimator__forward(self, inputs, *args, **kwargs):
    """Rewrite `forward` of TopDown for default backend.'.

    1.directly call _forward of subclass.

    Args:
        ctx (ContextCaller): The context with additional information.
        self (BasePoseEstimator): The instance of the class Object
            BasePoseEstimator.
        inputs (torch.Tensor[NxCxHxW]): Input images.

    Returns:
        torch.Tensor: The predicted heatmaps.
    """
    return self._forward(inputs)



# Copyright (c) OpenMMLab. All rights reserved.
# Modified from:
# https://github.com/pytorch/pytorch/blob/9ade03959392e5a90b74261012de1d806cab2253/torch/onnx/symbolic_opset9.py

from torch.onnx.symbolic_helper import parse_args

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER
from juxtapose.mmdeploy.utils import Backend


@SYMBOLIC_REWRITER.register_symbolic(
    "layer_norm", is_pytorch=True, arg_descriptors=["v", "is", "v", "v", "f", "i"]
)
def layer_norm__default(g, input, normalized_shape, weight, bias, eps, cudnn_enable):
    """Symbolic function for `layer_norm`

    Layer norm with torch<=1.12 might lead to wrong output shapes. Add
    keepdims=1 to each ReduceMean node to correct the shape.
    """
    import torch.onnx.symbolic_helper as sym_help
    from torch.onnx.symbolic_opset9 import add, mul, pow, sqrt, sub

    axes = [-i for i in range(len(normalized_shape), 0, -1)]

    two_cst = sym_help._generate_wrapped_number(g, 2.0)
    eps_cst = sym_help._generate_wrapped_number(g, eps)

    mean = g.op("ReduceMean", input, axes_i=axes, keepdims_i=1)
    numerator = sub(g, input, mean)
    # variance = e((x - e(x))^2), and (x - e(x)) is the numerator in the
    # layer_norm formula
    variance = g.op("ReduceMean", pow(g, numerator, two_cst), axes_i=axes, keepdims_i=1)
    denominator = sqrt(g, add(g, variance, eps_cst))

    layer_norm = g.op("Div", numerator, denominator)

    if not (weight is None or sym_help._is_none(weight)):
        layer_norm = mul(g, layer_norm, weight)
    if not (bias is None or sym_help._is_none(bias)):
        layer_norm = add(g, layer_norm, bias)

    return layer_norm


@parse_args("v", "is", "v", "v", "f", "i")
def _layer_norm_ncnn(g, input, normalized_shape, weight, bias, eps, cudnn_enable):
    """Symbolic function for `layer_norm`.

    PyTorch does not support export layer_norm to ONNX by default. We add the
    support here. `layer_norm` will be exported as ONNX node
    'mmdeploy::layer_norm'
    """
    weight.setDebugName("layernorm_weight")
    bias.setDebugName("layernorm_bias")
    return g.op("mmdeploy::LayerNorm", input, weight, bias, affine_i=1, epsilon_f=eps)


@SYMBOLIC_REWRITER.register_symbolic(
    "layer_norm", is_pytorch=True, backend=Backend.NCNN.value
)
def layer_norm__ncnn(*args):
    """Register default symbolic function for `layer_norm`.

    Add support to layer_norm to ONNX.
    """
    return _layer_norm_ncnn(*args)



# Copyright (c) OpenMMLab. All rights reserved.
import importlib

from juxtapose.mmdeploy.utils import Codebase


def get_library_version(lib):
    """Try to get the version of a library if it has been installed.

    Args:
        lib (str): The name of library.

    Returns:
        None | str: If the library has been installed, return version.
    """
    try:
        lib = importlib.import_module(lib)
        if hasattr(lib, "__version__"):
            version = lib.__version__
        else:
            version = None
    except Exception:
        version = None

    return version


def get_codebase_version():
    """Get the version dictionary of all supported codebases.

    Returns:
        Dict: The name and the version of supported codebases.
    """
    version_dict = dict()
    for enum in Codebase:
        codebase = enum.value
        version_dict[codebase] = get_library_version(codebase)
    return version_dict


def get_backend_version():
    """Get the version dictionary of some supported backend.

    Returns:
        Dict: The name and the version of some supported backend.
    """
    backend_library_list = ["tensorrt", "onnxruntime", "ncnn", "tvm"]
    version_dict = dict()
    for backend in backend_library_list:
        version_dict[backend] = get_library_version(backend)
    return version_dict



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Union

import torch
from packaging import version

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils.constants import Backend


# TODO add version control when MOD is supported by TensorRT
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.__mod__", backend=Backend.TENSORRT.value
)
def mod__tensorrt(
    input: torch.Tensor, other: Union[torch.Tensor, torch.NumberType], *args, **kwargs
) -> torch.Tensor:
    """Rewrite `mod` when exporting model to ONNX for TensorRT backend."""
    ctx = FUNCTION_REWRITER.get_context()
    if version.parse(torch.__version__) > version.parse("1.10.0"):
        return input - (input // other) * other
    return ctx.origin_func(input, other, *args, **kwargs)



from typing import Literal

DETECTOR_TYPES = Literal[
    "rtmdet-s",
    "rtmdet-m",
    "rtmdet-l",
    "groundingdino",
    "yolov8s",
    "yolov8m",
    "yolov8l",
]

POSE_ESTIMATOR_TYPES = Literal["rtmpose-s", "rtmpose-m", "rtmpose-l"]

TRACKER_TYPES = Literal["bytetrack", "botsort", ""]

DEVICE_TYPES = Literal["cpu", "cuda"]



# Copyright (c) OpenMMLab. All rights reserved.
from .detection_output import ncnn_detection_output_forward
from .prior_box import ncnn_prior_box_forward

__all__ = ['ncnn_detection_output_forward', 'ncnn_prior_box_forward']



# Copyright (c) OpenMMLab. All rights reserved.

from typing import Optional, Sequence, Union

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.normalize", backend="ncnn"
)
def normalize__ncnn(
    input: torch.Tensor, p: int = 2, dim: int = 1, eps: float = 1e-12, *args, **kwargs
):
    """Rewrite `normalize` for ncnn backend.

    Make sure L2 norm on channel dim and be exported to ncnn correctly.
    """
    ctx = FUNCTION_REWRITER.get_context()
    if dim < 0:
        dim += input.ndim
    assert dim != 0, "Should not normalize on batch index"
    origin_func = ctx.origin_func
    assert p == 2, "only support L2 norm"
    assert input.ndim in [3, 4]
    assert input.shape[0] == 1, f"only support batch size 1, but given {input.shape[0]}"
    if input.ndim == 3:
        output = (
            origin_func(input.transpose(1, dim).unsqueeze(2), p=p, dim=1, eps=eps)
            .squeeze(2)
            .transpose(1, dim)
        )
    else:
        # input.ndim == 4:
        if dim == 1:
            output = origin_func(input, p=p, dim=dim, eps=eps)
        else:
            output = origin_func(
                input.transpose(1, dim), p=p, dim=1, eps=eps
            ).transpose(1, dim)
    return output


@FUNCTION_REWRITER.register_rewriter(func_name="torch.norm", backend="ncnn")
def norm__ncnn(
    input: torch.Tensor,
    p: Optional[Union[int, str]] = "fro",
    dim: Optional[Union[int, Sequence]] = None,
    keepdim: Optional[bool] = False,
    out: Optional[torch.Tensor] = None,
    dtype: Optional[torch.dtype] = None,
):
    """Rewrite `torch.norm` for ncnn backend.

    Rewrite torch.norm when p is Frobenius norm to avoid FP16 exceed in ncnn
    Android platform.
    """
    ctx = FUNCTION_REWRITER.get_context()
    origin_func = ctx.origin_func
    if p == "fro" and (isinstance(dim, int) or len(dim) == 1):
        # Substitute Frobenius norm with L2 norm.
        return origin_func(input, p=2, dim=dim, keepdim=keepdim, out=out, dtype=dtype)
    else:
        return origin_func(input, p=p, dim=dim, keepdim=keepdim, out=out, dtype=dtype)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List

import torch
from mmdet.models.layers import mask_matrix_nms
from mmdet.utils import OptConfigType
from torch import Tensor
from torch.nn import functional as F

from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.solo_head." "SOLOHead.predict_by_feat"
)
def solohead__predict_by_feat(
    self,
    mlvl_mask_preds: List[Tensor],
    mlvl_cls_scores: List[Tensor],
    batch_img_metas: List[Dict],
    cfg: OptConfigType = None,
    **kwargs,
):
    """Rewrite `predict_by_feat` of `SOLOHead` for default backend."""

    ctx = FUNCTION_REWRITER.get_context()
    batch_size = mlvl_cls_scores[0].size(0)
    cfg = self.test_cfg
    mlvl_cls_scores = [
        item.permute(0, 2, 3, 1).view(item.size(0), -1, self.cls_out_channels)
        for item in mlvl_cls_scores
    ]

    # avoid setting items
    lvl_strides = [
        torch.ones_like(mlvl_cls_scores[lvl][0, :, 0]) * self.strides[lvl]
        for lvl in range(len(mlvl_cls_scores))
    ]
    strides = torch.cat(lvl_strides, 0)
    assert len(mlvl_mask_preds) == len(mlvl_cls_scores)
    batch_mlvl_cls_scores = torch.cat(mlvl_cls_scores, dim=1)
    batch_mlvl_mask_preds = torch.cat(mlvl_mask_preds, dim=1)
    featmap_size = batch_mlvl_mask_preds.size()[-2:]
    batch_mlvl_cls_scores, cls_labels = torch.max(batch_mlvl_cls_scores, -1)

    score_mask = batch_mlvl_cls_scores > cfg.score_thr
    # pad zero to filter items
    batch_mlvl_cls_scores = batch_mlvl_cls_scores.where(
        score_mask, batch_mlvl_cls_scores.new_zeros(1)
    ).view(-1)

    cls_labels = cls_labels.view(-1)

    mask_preds = batch_mlvl_mask_preds.view(-1, featmap_size[0], featmap_size[1])

    masks = mask_preds > cfg.mask_thr
    sum_masks = masks.sum((1, 2))
    keep = sum_masks > strides
    # pad zero to filter items
    cls_scores = batch_mlvl_cls_scores.where(keep, batch_mlvl_cls_scores.new_zeros(1))
    sum_masks = sum_masks.where(keep, sum_masks.new_ones(1))

    # maskness
    mask_scores = (mask_preds * masks).sum((1, 2)) / sum_masks
    cls_scores *= mask_scores
    sum_masks = sum_masks.where(keep, sum_masks.new_zeros(1))

    scores, labels, _, keep_inds = mask_matrix_nms(
        masks,
        cls_labels,
        cls_scores,
        mask_area=sum_masks,
        nms_pre=cfg.nms_pre,
        max_num=cfg.max_per_img,
        kernel=cfg.kernel,
        sigma=cfg.sigma,
        filter_thr=cfg.filter_thr,
    )

    h, w = batch_img_metas[0]["img_shape"][:2]
    mask_preds = mask_preds[keep_inds].unsqueeze(0)

    mmdet_params = get_post_processing_params(ctx.cfg)
    export_postprocess_mask = mmdet_params.get("export_postprocess_mask", True)
    if export_postprocess_mask:
        upsampled_size = (featmap_size[0] * 4, featmap_size[1] * 4)
        mask_preds = F.interpolate(mask_preds, size=upsampled_size, mode="bilinear")
        bboxes = scores.new_zeros(batch_size, scores.shape[-1], 4)
    else:
        bboxes = scores.new_zeros(batch_size, scores.shape[-1], 2)
        # full screen box so we can postprocess mask outside the model
        bboxes = torch.cat(
            [
                bboxes,
                bboxes.new_full((*bboxes.shape[:2], 1), w),
                bboxes.new_full((*bboxes.shape[:2], 1), h),
            ],
            dim=-1,
        )

    labels = labels.reshape(batch_size, -1)
    dets = torch.cat([bboxes, scores.reshape(batch_size, -1, 1)], dim=-1)

    return dets, labels, mask_preds



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Callable, Dict, Iterable, Optional

import onnx
from onnx.helper import get_attribute_value

from juxtapose.mmdeploy.utils import get_root_logger


def attribute_to_dict(attr: onnx.AttributeProto) -> Dict:
    """Convert onnx op attribute to dict.

    Args:
        attr (onnx.AttributeProto): Input onnx op attribute.

    Returns:
        dict: A dict contains info from op attribute.
    """
    ret = {}
    for a in attr:
        value = get_attribute_value(a)
        if isinstance(value, bytes):
            value = str(value, "utf-8")
        ret[a.name] = value
    return ret


def remove_nodes(model: onnx.ModelProto, predicate: Callable) -> onnx.ModelProto:
    """Remove nodes from ONNX model.

    Args:
        model (onnx.ModelProto): Input onnx model.
        predicate (Callable): A function to predicate a node.

    Returns:
        onnx.ModelProto: Modified onnx model.
    """
    # ! this doesn't handle inputs/outputs
    logger = get_root_logger()
    while True:
        connect = None
        for i, node in enumerate(model.graph.node):
            if predicate(node):
                assert len(node.input) == 1
                assert len(node.output) == 1
                connect = (node.input[0], node.output[0])
                logger.info(f"remove node {node.name}")
                del model.graph.node[i]
                break
        if not connect:
            break
        src, dst = connect
        for node in model.graph.node:
            for i, input in enumerate(node.input):
                if input == dst:
                    node.input[i] = src
    return model


def is_unused_mark(marks: Iterable[onnx.NodeProto]) -> Callable:
    """Check whether a mark is unused.

    Args:
        marks (Iterable[onnx.NodeProto]): A list of onnx NodeProto.

    Returns:
        Callable: The function to check if a mark node is in `marks`.
    """

    def f(node):
        if node.op_type == "Mark":
            attr = attribute_to_dict(node.attribute)
            name = attr["func"] + ":" + attr["type"]
            if name not in marks:
                return True
        return False

    return f


def is_identity(node: onnx.NodeProto) -> bool:
    """Check if an op is identity."""
    return node.op_type == "Identity"


def get_new_name(
    attrs: Dict[str, str],
    mark_name: str = "",
    name_map: Optional[Dict[str, str]] = None,
) -> str:
    """Get new name for a node.

    Args:
        attrs (Dict[str, str]): A dict contains attributes of an ONNX node.
        mark_name (str): The input mark op name. Default is ''.
        name_map (Dict[str, str]): A mapping of node names, defaults to
            `None`.

    Returns:
        str: The new node name.
    """
    if "name" in attrs:
        new_name = attrs["name"]
    else:
        new_name = "_".join((attrs["func"], attrs["type"], str(attrs["id"])))

    if name_map is not None:
        if new_name in name_map:
            return name_map[new_name]

        if f"{mark_name}:{new_name}" in name_map:
            return name_map[f"{mark_name}:{new_name}"]

    return new_name


def rename_value(model: onnx.ModelProto, old_name: str, new_name: str):
    """Rename a node in an ONNX model.

    Args:
        model (onnx.ModelProto): Input onnx model.
        old_name (str): Original node name in the model.
        new_name (str): New node name in the model.
    """
    if old_name == new_name:
        return
    logger = get_root_logger()
    logger.info(f"rename {old_name} -> {new_name}")
    for n in model.graph.node:
        for i, output in enumerate(n.output):
            if output == old_name:
                n.output[i] = new_name
        for i, input in enumerate(n.input):
            if input == old_name:
                n.input[i] = new_name
    for v in model.graph.value_info:
        if v.name == old_name:
            v.name = new_name
    for i, input in enumerate(model.graph.input):
        if input.name == old_name:
            input.name = new_name
    for i, output in enumerate(model.graph.output):
        if output.name == old_name:
            output.name = new_name


def remove_identity(model: onnx.ModelProto):
    """Remove identity node from an ONNX model.

    Args:
        model (onnx.ModelProto): Input onnx model.
    """
    graph = model.graph

    def simplify_inputs():
        connect = None
        logger = get_root_logger()
        for input in graph.input:
            for i, node in enumerate(graph.node):
                if node.op_type == "Identity" and node.input[0] == input.name:
                    connect = (node.input[0], node.output[0])
                    logger.info(f"remove node {node.name}")
                    del graph.node[i]
                    break
            if connect:
                break
        if not connect:
            return False
        src, dst = connect
        for node in graph.node:
            for i, input_name in enumerate(node.input):
                if input_name == dst:
                    node.input[i] = src
        # the input just changed won't be an output
        return True

    def simplify_outputs():
        connect = None
        logger = get_root_logger()
        for output in graph.output:
            for i, node in enumerate(graph.node):
                if node.op_type == "Identity" and node.output[0] == output.name:
                    connect = (node.input[0], node.output[0])
                    logger.info(f"remove node {node.name}")
                    del graph.node[i]
                    break
            if connect:
                break
        if not connect:
            return False
        src, dst = connect
        for node in graph.node:
            for i, output_name in enumerate(node.output):
                if output_name == src:
                    node.output[i] = dst
            # the output just renamed may be someone's input
            for i, input_name in enumerate(node.input):
                if input_name == src:
                    node.input[i] = dst
        return True

    while simplify_inputs():
        pass

    while simplify_outputs():
        pass

    remove_nodes(model, is_identity)


def remove_imports(model: onnx.ModelProto):
    """Remove useless imports from an ONNX model.

    The domain like `mmdeploy` might influence model conversion for
    some backends.

    Args:
        model (onnx.ModelProto): Input onnx model.
    """
    logger = get_root_logger()
    dst_domain = [""]
    for node in model.graph.node:
        if hasattr(node, "module") and (node.module not in dst_domain):
            dst_domain.append(node.module)
    src_domains = [oi.domain for oi in model.opset_import]
    for i, src_domain in enumerate(src_domains):
        if src_domain not in dst_domain:
            logger.info(f"remove opset_import {src_domain}")
            model.opset_import.pop(i)



# Copyright (c) OpenMMLab. All rights reserved.
from . import adaptive_pool  # noqa: F401,F403
from . import any  # noqa: F401,F403
from . import atan2  # noqa: F401,F403
from . import cat  # noqa: F401,F403
from . import chunk  # noqa: F401,F403
from . import clip  # noqa: F401,F403
from . import expand  # noqa: F401,F403
from . import flatten  # noqa: F401,F403
from . import getattribute  # noqa: F401,F403
from . import group_norm  # noqa: F401,F403
from . import interpolate  # noqa: F401,F403
from . import linear  # noqa: F401,F403
from . import linspace  # noqa: F401,F403
from . import masked_fill  # noqa: F401,F403
from . import mod  # noqa: F401,F403
from . import multi_head_attention_forward  # noqa: F401,F403
from . import normalize  # noqa: F401,F403
from . import pad  # noqa: F401,F403
from . import repeat  # noqa: F401,F403
from . import size  # noqa: F401,F403
from . import tensor_getitem  # noqa: F401,F403
from . import tensor_setitem  # noqa: F401,F403
from . import topk  # noqa: F401,F403
from . import triu  # noqa: F401,F403



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import Any, Optional, Union

import mmengine

from .core import PIPELINE_MANAGER


@PIPELINE_MANAGER.register_pipeline()
def torch2onnx(
    img: Any,
    work_dir: str,
    save_file: str,
    deploy_cfg: Union[str, mmengine.Config],
    model_cfg: Union[str, mmengine.Config],
    model_checkpoint: Optional[str] = None,
    device: str = "cuda:0",
):
    """Convert PyTorch model to ONNX model.

    Examples:
        >>> from juxtapose.mmdeploy.apis import torch2onnx
        >>> img = 'demo.jpg'
        >>> work_dir = 'work_dir'
        >>> save_file = 'fcos.onnx'
        >>> deploy_cfg = ('configs/mmdet/detection/'
                          'detection_onnxruntime_dynamic.py')
        >>> model_cfg = ('mmdetection/configs/fcos/'
                         'fcos_r50_caffe_fpn_gn-head_1x_coco.py')
        >>> model_checkpoint = ('checkpoints/'
                                'fcos_r50_caffe_fpn_gn-head_1x_coco-821213aa.pth')
        >>> device = 'cpu'
        >>> torch2onnx(img, work_dir, save_file, deploy_cfg, \
            model_cfg, model_checkpoint, device)

    Args:
        img (str | np.ndarray | torch.Tensor): Input image used to assist
            converting model.
        work_dir (str): A working directory to save files.
        save_file (str): Filename to save onnx model.
        deploy_cfg (str | mmengine.Config): Deployment config file or
            Config object.
        model_cfg (str | mmengine.Config): Model config file or Config object.
        model_checkpoint (str): A checkpoint path of PyTorch model,
            defaults to `None`.
        device (str): A string specifying device type, defaults to 'cuda:0'.
    """

    from juxtapose.mmdeploy.apis.core.pipeline_manager import no_mp
    from juxtapose.mmdeploy.utils import (
        Backend,
        get_backend,
        get_dynamic_axes,
        get_input_shape,
        get_onnx_config,
        load_config,
    )
    from .onnx import export

    # load deploy_cfg if necessary
    deploy_cfg, model_cfg = load_config(deploy_cfg, model_cfg)
    mmengine.mkdir_or_exist(osp.abspath(work_dir))

    input_shape = get_input_shape(deploy_cfg)

    # create model an inputs
    from juxtapose.mmdeploy.apis import build_task_processor

    task_processor = build_task_processor(model_cfg, deploy_cfg, device)

    torch_model = task_processor.build_pytorch_model(model_checkpoint)
    data, model_inputs = task_processor.create_input(
        img,
        input_shape,
        data_preprocessor=getattr(torch_model, "data_preprocessor", None),
    )

    if isinstance(model_inputs, list) and len(model_inputs) == 1:
        model_inputs = model_inputs[0]
    data_samples = data["data_samples"]
    input_metas = {"data_samples": data_samples, "mode": "predict"}

    # export to onnx
    context_info = dict()
    context_info["deploy_cfg"] = deploy_cfg
    output_prefix = osp.join(work_dir, osp.splitext(osp.basename(save_file))[0])
    backend = get_backend(deploy_cfg).value

    onnx_cfg = get_onnx_config(deploy_cfg)
    opset_version = onnx_cfg.get("opset_version", 11)

    input_names = onnx_cfg["input_names"]
    output_names = onnx_cfg["output_names"]
    axis_names = input_names + output_names
    dynamic_axes = get_dynamic_axes(deploy_cfg, axis_names)
    verbose = not onnx_cfg.get("strip_doc_string", True) or onnx_cfg.get(
        "verbose", False
    )
    keep_initializers_as_inputs = onnx_cfg.get("keep_initializers_as_inputs", True)
    optimize = onnx_cfg.get("optimize", False)
    if backend == Backend.NCNN.value:
        """NCNN backend needs a precise blob counts, while using onnx optimizer
        will merge duplicate initilizers without reference count."""
        optimize = False
    with no_mp():
        export(
            torch_model,
            model_inputs,
            input_metas=input_metas,
            output_path_prefix=output_prefix,
            backend=backend,
            input_names=input_names,
            output_names=output_names,
            context_info=context_info,
            opset_version=opset_version,
            dynamic_axes=dynamic_axes,
            verbose=verbose,
            keep_initializers_as_inputs=keep_initializers_as_inputs,
            optimize=optimize,
        )



# Copyright (c) OpenMMLab. All rights reserved.
from collections import defaultdict
from typing import Callable, Dict, List, Optional, Sequence, Union

import torch
from torch.autograd import Function
from torch.onnx.symbolic_helper import parse_args

from juxtapose.mmdeploy.utils import IR, Backend, get_root_logger
from .rewriter_utils import (
    Checker,
    ContextCaller,
    RewriterRegistry,
    copy_function,
    eval_with_import,
    get_frame_func,
    get_func_qualname,
)


class SymbolicRewriter:
    """A symbolic rewriter which maintains rewritten symbolic.

    The rewritten symbolic can be registered by calling register_symbolic(). In
    RewriteContext, the rewriter automatically registers extra symbolic of
    pytorch and replaces symbolic of custom functions. The symbolic will
    recover after exiting the RewriteContext.

    Examples:
        >>> @SYMBOLIC_REWRITER.register_symbolic('squeeze', \
        >>> is_pytorch=True)
        >>> def squeeze_default(g, self, dim=None):
        >>>     if dim is None:
        >>>         dims = []
        >>>         for i, size in enumerate(self.type().sizes()):
        >>>             if size == 1:
        >>>                 dims.append(i)
        >>>     else:
        >>>         dims = [sym_help._get_const(dim, 'i', 'dim')]
        >>>     return g.op('Squeeze', self, axes_i=dims)
    """

    def __init__(self) -> None:
        self._registry = RewriterRegistry()
        self._func_contexts = defaultdict(list)

    def register_symbolic(
        self,
        func_name: str,
        backend: str = Backend.DEFAULT.value,
        is_pytorch: bool = False,
        arg_descriptors: Optional[Sequence[str]] = None,
        ir: IR = IR.DEFAULT,
        extra_checkers: Optional[Union[Checker, List[Checker]]] = None,
        **kwargs,
    ) -> Callable:
        """The decorator of the custom symbolic.

        Args:
            func_name (str): The function name/path to override the symbolic.
            backend (str): The rewriter will be activated on which backend.
            is_pytorch (bool): Enable this flag if func_name is the name of \
                a pytorch builtin function.
            arg_descriptors (Sequence[str]): The argument descriptors of the \
                symbol.
            ir (IR): The rewriter will be activated on which IR.
            extra_checkers (Checker | List[Checker] | None): Other requirements
                defined by Checker.

        Returns:
            Callable: The process of registered symbolic.
        """
        return self._registry.register_object(
            func_name,
            backend,
            ir,
            extra_checkers,
            is_pytorch=is_pytorch,
            arg_descriptors=arg_descriptors,
            **kwargs,
        )

    def enter(self, cfg: Dict = dict(), env: Dict = dict(), opset: int = 11, **kwargs):
        """The implementation of symbolic register."""
        # clear context
        self._func_contexts.clear()

        # Get current records
        symbolic_records = self._registry.get_records(env)

        self._pytorch_symbolic = list()
        self._extra_symbolic = list()
        new_functions = list()
        for function_name, record_dict in symbolic_records:
            symbolic_function = record_dict["_object"]
            symbolic_function = copy_function(symbolic_function)
            arg_descriptors = record_dict["arg_descriptors"]
            extra_kwargs = kwargs.copy()
            extra_kwargs.update(record_dict)
            context_caller = ContextCaller(symbolic_function, None, cfg, **extra_kwargs)

            # register context
            qualname = get_func_qualname(symbolic_function)
            self._func_contexts[qualname].append(context_caller)
            self._func_contexts[function_name].append(context_caller)

            if arg_descriptors is not None and len(arg_descriptors) > 0:
                symbolic_function = parse_args(*arg_descriptors)(symbolic_function)

            is_pytorch = record_dict["is_pytorch"]
            if is_pytorch:
                from torch.onnx import register_custom_op_symbolic

                register_custom_op_symbolic(
                    f"::{function_name}", symbolic_function, opset
                )

                # Save domain and version
                self._pytorch_symbolic.append((function_name, "", opset))
            else:
                # Check if the origin function exists
                try:
                    origin_func = eval_with_import(function_name)
                    assert issubclass(
                        origin_func, Function
                    ), f"{function_name} is not an torch.autograd.Function"
                except Exception:
                    origin_func = None
                    logger = get_root_logger()
                    logger.warning(f"Can not add symbolic for `{function_name}`")

                # Only register functions that exist
                if origin_func is not None:
                    origin_symbolic = getattr(origin_func, "symbolic", None)

                    # Save origin function
                    self._extra_symbolic.append((origin_func, origin_symbolic))

                    # Cache new the function to avoid homonymic bug
                    new_functions.append((origin_func, symbolic_function))

            for origin_func, new_func in new_functions:
                origin_symbolic = getattr(origin_func, "symbolic", None)
                new_func.origin_func = origin_symbolic
                origin_func.symbolic = new_func

    def exit(self):
        """The implementation of symbolic unregister."""
        # clear context
        self._func_contexts.clear()

        # Unregister pytorch op
        if hasattr(torch.onnx, "unregister_custom_op_symbolic"):
            from torch.onnx import unregister_custom_op_symbolic

            for function_name, domain, version in self._pytorch_symbolic:
                unregister_custom_op_symbolic(f"::{function_name}", version)
        else:
            from torch.onnx.symbolic_registry import _registry as pytorch_registry

            for function_name, domain, version in self._pytorch_symbolic:
                # Same to unregister_op() in torch 1.9.0+
                del pytorch_registry[(domain, version)][function_name]
                if not pytorch_registry[(domain, version)]:
                    del pytorch_registry[(domain, version)]

        # Unregister custom op
        for origin_func, origin_symbolic in self._extra_symbolic:
            origin_func.symbolic = origin_symbolic

    def get_context(self, key: Optional[str] = None) -> ContextCaller:
        """Get the context of rewriter.

        Args:
            key: key to the context.

        Returns:
            ContextCaller: context of function
        """
        func = None
        if key is None:
            func = get_frame_func(2)
            key = get_func_qualname(func)

        # get all contexts
        ctxs = self._func_contexts.get(key, [])

        if func is None:
            assert len(ctxs) == 1
            return ctxs[0]

        ctx = None
        for tmp_ctx in ctxs:
            if tmp_ctx.func == func:
                ctx = tmp_ctx

        if ctx is None:
            get_root_logger().warning(f"Can not found context of {key}")
        return ctx



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Sequence

import torch
from mmengine.config import ConfigDict
from mmengine.structures import InstanceData
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import (
    gather_topk,
    get_post_processing_params,
    pad_with_value_if_necessary,
)
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms
from juxtapose.mmdeploy.utils import is_dynamic_shape


def _bbox_pre_decode(
    points: torch.Tensor, bbox_pred: torch.Tensor, stride: torch.Tensor
):
    """compute real bboxes."""
    points = points[..., :2]
    bbox_pos_center = torch.cat([points, points], dim=-1)
    bboxes = bbox_pred * stride + bbox_pos_center
    return bboxes


def _bbox_post_decode(bboxes: torch.Tensor, max_shape: Sequence[int]):
    """clamp bbox."""
    x1 = bboxes[..., 0].clamp(min=0, max=max_shape[1])
    y1 = bboxes[..., 1].clamp(min=0, max=max_shape[0])
    x2 = bboxes[..., 2].clamp(min=0, max=max_shape[1])
    y2 = bboxes[..., 3].clamp(min=0, max=max_shape[0])
    decoded_bboxes = torch.stack([x1, y1, x2, y2], dim=-1)
    return decoded_bboxes


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.dense_heads.reppoints_head.RepPointsHead.points2bbox"
)
def reppoints_head__points2bbox(self, pts, y_first=True):
    """Rewrite of `points2bbox` in `RepPointsHead`.

    Use `self.moment_transfer` in `points2bbox` will cause error:
    RuntimeError: Input, output and indices must be on the current device
    """
    ctx = FUNCTION_REWRITER.get_context()
    update_moment = hasattr(self, "moment_transfer")
    if update_moment:
        moment_transfer = self.moment_transfer
        delattr(self, "moment_transfer")
        self.moment_transfer = torch.tensor(moment_transfer.data)
    ret = ctx.origin_func(self, pts, y_first=y_first)
    if update_moment:
        self.moment_transfer = moment_transfer
    return ret


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.dense_heads.reppoints_head.RepPointsHead.predict_by_feat"
)
def reppoints_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    score_factors: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = False,
    with_nms: bool = True,
) -> InstanceData:
    """Rewrite `predict_by_feat` of `RepPointsHead` for default backend.

    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.

    Args:
        ctx (ContextCaller): The context with additional information.
        self (RepPointsHead): The instance of the class RepPointsHead.
        cls_scores (list[Tensor]): Box scores for each scale level
            with shape (N, num_anchors * num_classes, H, W).
        bbox_preds (list[Tensor]): Box energies / deltas for each scale
            level with shape (N, num_anchors * 4, H, W).
        score_factors (list[Tensor], Optional): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, num_priors * 1, H, W). Default None.
        img_metas (list[dict]):  Meta information of the image, e.g.,
            image size, scaling factor, etc.
        cfg (mmengine.Config | None): Test / postprocessing configuration,
            if None, test_cfg would be used. Default: None.
        rescale (bool): If True, return boxes in original image space.
            Default: False.

    Returns:
        tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
    """
    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg
    is_dynamic_flag = is_dynamic_shape(deploy_cfg)
    num_levels = len(cls_scores)

    featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
    mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, dtype=bbox_preds[0].dtype, device=bbox_preds[0].device
    )
    mlvl_priors = [priors.unsqueeze(0) for priors in mlvl_priors]

    mlvl_cls_scores = [cls_scores[i].detach() for i in range(num_levels)]
    mlvl_bbox_preds = [bbox_preds[i].detach() for i in range(num_levels)]
    assert batch_img_metas is not None
    img_shape = batch_img_metas[0]["img_shape"]

    assert len(cls_scores) == len(bbox_preds) == len(mlvl_priors)
    batch_size = cls_scores[0].shape[0]
    cfg = self.test_cfg
    pre_topk = cfg.get("nms_pre", -1)

    mlvl_valid_bboxes = []
    mlvl_valid_scores = []

    for level_idx, (cls_score, bbox_pred, priors) in enumerate(
        zip(mlvl_cls_scores, mlvl_bbox_preds, mlvl_priors)
    ):
        assert cls_score.size()[-2:] == bbox_pred.size()[-2:]

        scores = cls_score.permute(0, 2, 3, 1).reshape(
            batch_size, -1, self.cls_out_channels
        )
        if self.use_sigmoid_cls:
            scores = scores.sigmoid()
        else:
            scores = scores.softmax(-1)
        bbox_pred = bbox_pred.permute(0, 2, 3, 1)
        bbox_pred = bbox_pred.reshape(batch_size, -1)
        bbox_pred = (bbox_pred + 0).reshape(batch_size, -1, 4)
        if not is_dynamic_flag:
            priors = priors.data
        if pre_topk > 0:
            priors = pad_with_value_if_necessary(priors, 1, pre_topk)
            bbox_pred = pad_with_value_if_necessary(bbox_pred, 1, pre_topk)
            scores = pad_with_value_if_necessary(scores, 1, pre_topk, 0.0)

            nms_pre_score = scores

            # Get maximum scores for foreground classes.
            if self.use_sigmoid_cls:
                max_scores, _ = nms_pre_score.max(-1)
            else:
                max_scores, _ = nms_pre_score[..., :-1].max(-1)
            _, topk_inds = max_scores.topk(pre_topk)
            bbox_pred, scores = gather_topk(
                bbox_pred,
                scores,
                inds=topk_inds,
                batch_size=batch_size,
                is_batched=True,
            )
            priors = gather_topk(
                priors, inds=topk_inds, batch_size=batch_size, is_batched=False
            )

        bbox_pred = _bbox_pre_decode(priors, bbox_pred, self.point_strides[level_idx])
        mlvl_valid_bboxes.append(bbox_pred)
        mlvl_valid_scores.append(scores)

    batch_mlvl_bboxes_pred = torch.cat(mlvl_valid_bboxes, dim=1)
    batch_scores = torch.cat(mlvl_valid_scores, dim=1)
    batch_bboxes = _bbox_post_decode(bboxes=batch_mlvl_bboxes_pred, max_shape=img_shape)

    if cfg.get("min_bbox_size", -1) >= 0:
        w = batch_bboxes[:, :, 2] - batch_bboxes[:, :, 0]
        h = batch_bboxes[:, :, 3] - batch_bboxes[:, :, 1]
        valid_mask = (w > cfg.min_bbox_size) & (h > cfg.min_bbox_size)
        if not valid_mask.all():
            batch_scores = batch_scores * valid_mask.unsqueeze(-1)

    if not self.use_sigmoid_cls:
        batch_scores = batch_scores[..., : self.num_classes]

    post_params = get_post_processing_params(deploy_cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.pre_top_k
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
    nms_type = cfg.nms.get("type")
    return multiclass_nms(
        batch_bboxes,
        batch_scores,
        max_output_boxes_per_class,
        nms_type=nms_type,
        iou_threshold=iou_threshold,
        score_threshold=score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
    )



import json
import time


class TimeCounter:
    def __init__(self) -> None:
        pass

    def clear(self):
        self.timedict = {}
        self.basetime = time.perf_counter()

    def timeit(self, name):
        nowtime = time.perf_counter() - self.basetime
        self.timedict[name] = nowtime
        self.basetime = time.perf_counter()


class TimeHolder:
    def __init__(self) -> None:
        self.timedict = {}

    def update(self, _timedict: dict):
        for k, v in _timedict.items():
            if k not in self.timedict:
                self.timedict[k] = AverageMeter(name=k, val_only=True)
            self.timedict[k].update(val=v)

    def final_res(self):
        return {k: v.avg for k, v in self.timedict.items()}

    def __str__(self):
        return json.dumps(self.final_res(), indent=2)


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self, name, fmt=":f", val_only=False):
        self.name = name
        self.fmt = fmt
        self.val_only = val_only
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        if self.val_only:
            fmtstr = "{name} {val" + self.fmt + "}"
        else:
            fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER, SYMBOLIC_REWRITER
from juxtapose.mmdeploy.utils import IR


@FUNCTION_REWRITER.register_rewriter(
    "mmcv.ops.modulated_deform_conv.modulated_deform_conv2d", ir=IR.TORCHSCRIPT
)
def modulated_deform_conv__torchscript(
    input, offset, mask, weight, bias, stride, padding, dilation, groups, deform_groups
):
    """rewriter for the custom torchscript mdcn op."""
    from juxtapose.mmdeploy.backend.torchscript import get_ops_path, ops_available

    assert ops_available(), "torchscript custom ops is required."
    torch.ops.load_library(get_ops_path())
    from torch.nn.modules.utils import _pair

    kernel_h, kernel_w = weight.shape[-2:]
    stride = _pair(stride)
    padding = _pair(padding)
    dilation = _pair(dilation)
    with_bias = bias is not None
    if not with_bias:
        bias = input.new_empty(0)
    return torch.ops.mmdeploy.modulated_deform_conv(
        input,
        weight,
        bias,
        offset,
        mask,
        kernel_h,
        kernel_w,
        stride[1],
        stride[0],
        padding[1],
        padding[0],
        dilation[1],
        dilation[0],
        groups,
        deform_groups,
        with_bias,
    )


@SYMBOLIC_REWRITER.register_symbolic(
    "mmcv.ops.modulated_deform_conv.ModulatedDeformConv2dFunction"
)
def modulated_deform_conv_default(
    g,
    input,
    offset,
    mask,
    weight,
    bias,
    stride,
    padding,
    dilation,
    groups,
    deform_groups,
):
    """Rewrite mdcn symbolic function for all backend."""
    input_tensors = [input, offset, mask, weight]
    if bias is not None:
        input_tensors.append(bias)
    return g.op(
        "mmdeploy::MMCVModulatedDeformConv2d",
        *input_tensors,
        stride_i=stride,
        padding_i=padding,
        dilation_i=dilation,
        groups_i=groups,
        deform_groups_i=deform_groups,
    )



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List

import torch
import torch.nn.functional as F
from mmdet.models.layers.matrix_nms import mask_matrix_nms
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.solov2_head." "SOLOV2Head.predict_by_feat"
)
def solov2_head__predict_by_feat(
    self,
    mlvl_kernel_preds: List[Tensor],
    mlvl_cls_scores: List[Tensor],
    mask_feats: Tensor,
    batch_img_metas: List[Dict],
    **kwargs,
):
    """Rewrite `predict_by_feat` of `SOLOV2Head` for default backend.

    Args:
        mlvl_kernel_preds (list[Tensor]): Multi-level dynamic kernel
            prediction. The kernel is used to generate instance
            segmentation masks by dynamic convolution. Each element in the
            list has shape
            (batch_size, kernel_out_channels, num_grids, num_grids).
        mlvl_cls_scores (list[Tensor]): Multi-level scores. Each element
            in the list has shape
            (batch_size, num_classes, num_grids, num_grids).
        mask_feats (Tensor): Unified mask feature map used to generate
            instance segmentation masks by dynamic convolution. Has shape
            (batch_size, mask_out_channels, h, w).
        batch_img_metas (list[dict]): Meta information of all images.

    Returns:
        list[:obj:`InstanceData`]: Processed results of multiple
        images.Each :obj:`InstanceData` usually contains
        following keys.

            - scores (Tensor): Classification scores, has shape
                (num_instance,).
            - labels (Tensor): Has shape (num_instances,).
            - masks (Tensor): Processed mask results, has
                shape (num_instances, h, w).
    """
    ctx = FUNCTION_REWRITER.get_context()
    cfg = self.test_cfg
    num_levels = len(mlvl_cls_scores)
    batch_size = mlvl_cls_scores[0].size(0)
    assert len(mlvl_kernel_preds) == len(mlvl_cls_scores)

    for lvl in range(num_levels):
        kernel_preds = mlvl_kernel_preds[lvl]
        cls_scores = mlvl_cls_scores[lvl]
        cls_scores = cls_scores.sigmoid()
        local_max = F.max_pool2d(cls_scores, 2, stride=1, padding=1)
        keep_mask = local_max[:, :, :-1, :-1] == cls_scores
        cls_scores = cls_scores * keep_mask
        mlvl_cls_scores[lvl] = cls_scores.permute(0, 2, 3, 1).view(
            batch_size, -1, self.cls_out_channels
        )
        mlvl_kernel_preds[lvl] = kernel_preds.permute(0, 2, 3, 1).view(
            batch_size, -1, self.kernel_out_channels
        )

    # Rewrite strides to avoid set_items.
    mlvl_strides = [
        torch.ones_like(mlvl_cls_scores[lvl][0, :, 0]) * self.strides[lvl]
        for lvl in range(len(mlvl_cls_scores))
    ]
    strides = torch.cat(mlvl_strides, 0)
    assert len(mlvl_kernel_preds) == len(mlvl_cls_scores)
    batch_mlvl_cls_scores = torch.cat(mlvl_cls_scores, dim=1)
    batch_mlvl_kernel_preds = torch.cat(mlvl_kernel_preds, dim=1)

    featmap_size = mask_feats.size()[-2:]
    h, w = batch_img_metas[0]["img_shape"][:2]
    batch_mlvl_cls_scores, cls_labels = torch.max(batch_mlvl_cls_scores, -1)

    score_mask = batch_mlvl_cls_scores > cfg.score_thr
    batch_mlvl_cls_scores = batch_mlvl_cls_scores.where(
        score_mask, batch_mlvl_cls_scores.new_zeros(1)
    ).view(-1)

    cls_labels = cls_labels.view(-1)

    # mask encoding.

    kernel_preds = batch_mlvl_kernel_preds[0].unsqueeze(2).unsqueeze(3)
    mask_preds = F.conv2d(mask_feats, kernel_preds, stride=1).squeeze(0).sigmoid()
    aligned_score_mask = score_mask[0].unsqueeze(1).unsqueeze(2)
    mask_preds = mask_preds.where(aligned_score_mask, mask_preds.new_zeros(1))

    # mask.
    masks = mask_preds > cfg.mask_thr
    sum_masks = masks.sum((1, 2))
    keep = sum_masks > strides
    cls_scores = batch_mlvl_cls_scores.where(keep, batch_mlvl_cls_scores.new_zeros(1))
    sum_masks = sum_masks.where(keep, sum_masks.new_ones(1))

    # maskness.
    mask_scores = (mask_preds * masks).sum((1, 2)) / sum_masks
    cls_scores *= mask_scores
    sum_masks = sum_masks.where(keep, sum_masks.new_zeros(1))

    scores, labels, _, keep_inds = mask_matrix_nms(
        masks,
        cls_labels,
        cls_scores,
        mask_area=sum_masks,
        nms_pre=cfg.nms_pre,
        max_num=cfg.max_per_img,
        kernel=cfg.kernel,
        sigma=cfg.sigma,
        filter_thr=cfg.filter_thr,
    )

    mask_preds = mask_preds[keep_inds].unsqueeze(0)
    post_params = get_post_processing_params(ctx.cfg)
    export_postprocess_mask = post_params.get("export_postprocess_mask", True)
    if export_postprocess_mask:
        upsampled_size = (
            featmap_size[0] * self.mask_stride,
            featmap_size[1] * self.mask_stride,
        )
        mask_preds = F.interpolate(mask_preds, size=upsampled_size, mode="bilinear")
        bboxes = scores.new_zeros(batch_size, scores.shape[-1], 4)
    else:
        bboxes = scores.new_zeros(batch_size, scores.shape[-1], 2)
        # full screen box so we can postprocess mask outside the model
        bboxes = torch.cat(
            [
                bboxes,
                bboxes.new_full((*bboxes.shape[:2], 1), w),
                bboxes.new_full((*bboxes.shape[:2], 1), h),
            ],
            dim=-1,
        )

    labels = labels.reshape(batch_size, -1)
    dets = torch.cat([bboxes, scores.reshape(batch_size, -1, 1)], dim=-1)

    return dets, labels, mask_preds



# Copyright (c) OpenMMLab. All rights reserved.
from enum import Enum


class AdvancedEnum(Enum):
    """Define an enumeration class."""

    @classmethod
    def get(cls, value):
        """Get the key through a value."""
        for k in cls:
            if k.value == value:
                return k

        raise KeyError(f'Cannot get key by value "{value}" of {cls}')


class Task(AdvancedEnum):
    """Define task enumerations."""
    TEXT_DETECTION = 'TextDetection'
    TEXT_RECOGNITION = 'TextRecognition'
    SEGMENTATION = 'Segmentation'
    SUPER_RESOLUTION = 'SuperResolution'
    CLASSIFICATION = 'Classification'
    OBJECT_DETECTION = 'ObjectDetection'
    INSTANCE_SEGMENTATION = 'InstanceSegmentation'
    VOXEL_DETECTION = 'VoxelDetection'
    POSE_DETECTION = 'PoseDetection'
    ROTATED_DETECTION = 'RotatedDetection'
    VIDEO_RECOGNITION = 'VideoRecognition'
    ModelCompress = 'ModelCompress'


class Codebase(AdvancedEnum):
    """Define codebase enumerations."""
    MMDET = 'mmdet'
    MMSEG = 'mmseg'
    MMPRETRAIN = 'mmpretrain'
    MMOCR = 'mmocr'
    MMAGIC = 'mmagic'
    MMDET3D = 'mmdet3d'
    MMPOSE = 'mmpose'
    MMROTATE = 'mmrotate'
    MMACTION = 'mmaction'
    MMRAZOR = 'mmrazor'
    MMYOLO = 'mmyolo'


class IR(AdvancedEnum):
    """Define intermediate representation enumerations."""
    ONNX = 'onnx'
    TORCHSCRIPT = 'torchscript'
    DEFAULT = 'default'


class Backend(AdvancedEnum):
    """Define backend enumerations."""
    PYTORCH = 'pytorch'
    TENSORRT = 'tensorrt'
    ONNXRUNTIME = 'onnxruntime'
    PPLNN = 'pplnn'
    NCNN = 'ncnn'
    SNPE = 'snpe'
    OPENVINO = 'openvino'
    SDK = 'sdk'
    TORCHSCRIPT = 'torchscript'
    RKNN = 'rknn'
    ASCEND = 'ascend'
    COREML = 'coreml'
    TVM = 'tvm'
    VACC = 'vacc'
    DEFAULT = 'default'


SDK_TASK_MAP = {
    Task.CLASSIFICATION:
    dict(component='LinearClsHead', cls_name='Classifier'),
    Task.OBJECT_DETECTION:
    dict(component='ResizeBBox', cls_name='Detector'),
    Task.INSTANCE_SEGMENTATION:
    dict(component='ResizeInstanceMask', cls_name='Detector'),
    Task.SEGMENTATION:
    dict(component='ResizeMask', cls_name='Segmentor'),
    Task.SUPER_RESOLUTION:
    dict(component='TensorToImg', cls_name='Restorer'),
    Task.TEXT_DETECTION:
    dict(component='TextDetHead', cls_name='TextDetector'),
    Task.TEXT_RECOGNITION:
    dict(component='CTCConvertor', cls_name='TextRecognizer'),
    Task.POSE_DETECTION:
    dict(component='Detector', cls_name='PoseDetector'),
    Task.ROTATED_DETECTION:
    dict(component='ResizeRBBox', cls_name='RotatedDetector'),
    Task.VIDEO_RECOGNITION:
    dict(component='BaseHead', cls_name='VideoRecognizer')
}

TENSORRT_MAX_TOPK = 3840



# Copyright (c) OpenMMLab. All rights reserved.
from .backend_model import BaseBackendModel
from .mmcodebase import CODEBASE, MMCodebase, get_codebase_class
from .task import BaseTask

__all__ = [
    'BaseBackendModel', 'BaseTask', 'MMCodebase', 'get_codebase_class',
    'CODEBASE'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

from torch import Tensor

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER


# Here using mmcv.ops.roi_align_rotated.__self__ to find
# mmcv.ops.roi_align.RoIAlignRotatedFunction, because RoIAlignRotatedFunction
# is not visible in mmcv.
@SYMBOLIC_REWRITER.register_symbolic(
    "mmcv.ops.roi_align_rotated.__self__", backend="default"
)
def roi_align_rotated_default(
    g,
    input: Tensor,
    rois: Tensor,
    output_size: List[int],
    spatial_scale: float,
    sampling_ratio: int,
    aligned: bool,
    clockwise: bool,
):
    """Rewrite symbolic function for default backend.

    Replace onnx::RoIAlignRotated with mmdeploy::MMCVRoIAlignRotated.

    Args:
        ctx (ContextCaller): The context with additional information.
        g (Graph): The traced onnx graph.
        input (Tensor): Input tensor, 4-D feature map of shape (N, C, H, W).
        rois (Tensor): Bx5 boxes. First column is the index into N. The other
            4 columns are xyxy.
        output_size(List[int]): Output size of height and width.
        spatial_scale (float):
        sampling_ratio (int): Number of inputs samples to take for each
            output sample. 0 to take samples densely for current models.
        aligned (bool): With `aligned=True`, we first appropriately scale
            the ROI and then shift it by -0.5 prior to calling roi_align.
            This produces the correct neighbors;
        clockwise (bool): If True, the angle in each proposal follows a
            clockwise fashion in image space, otherwise, the angle is
            counterclockwise. Default: False.

    Returns:
        MMCVRoiAlign op for onnx.
    """
    return g.op(
        "mmdeploy::MMCVRoIAlignRotated",
        input,
        rois,
        output_height_i=output_size[0],
        output_width_i=output_size[1],
        spatial_scale_f=spatial_scale,
        sampling_ratio_i=sampling_ratio,
        aligned_i=aligned,
        clockwise_i=clockwise,
    )



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.onnx.symbolic_helper as sym_help
from packaging.version import parse as version_parse

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


# modified from
# https://github.com/pytorch/pytorch/blob/65a37923f9b14c7c9e80535d771ef9e4e92d0502/torch/onnx/symbolic_opset11.py
@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.onnx.symbolic_opset11._prepare_onnx_paddings", backend="tensorrt"
)
def _prepare_onnx_paddings__tensorrt(g, input, pad):
    """Rewrite `_prepare_onnx_paddings` for TensorRT backend.

    For codes like `x = torch.nn.ZeroPad2d((0, a, 0, b))(x)`, where a and b are
    variables of torch.tensor, onnx2tensorrt raises errors like
    `INVALID_NODE: Invalid Node - Pad_`.

    Generate paddings in ONNX order based on pad in pytorch.
    Args:
        input: the input tensor.
        pad: the paddings in pytorch.
            The order is dim_n_begin, dim_n_end, dim_n-1_begin, dim_n-1_end,
            ..., dim_m_begin, dim_m_end,
            where m is in range [0, n].
    """
    ctx = FUNCTION_REWRITER.get_context()
    torch_version = version_parse(torch.__version__)
    if torch_version.major == 1 and torch_version.minor < 10:
        return ctx.origin_func(g, input, pad)
    # The desired order of paddings is
    # dim_0_begin, dim_1_begin, ... , dim_0_end, ..., dim_n_end.
    # n is the dimension of input.
    # Assume zero-dimensions in the beginning, pad the "pad" sequence with
    # zeros in the beginning
    pad_len = torch.onnx.symbolic_opset9.size(
        g, pad, g.op("Constant", value_t=torch.tensor([0]))
    )
    # Set extension = [0] * (dim * 2 - len(pad))
    rank = sym_help._get_tensor_rank(input)
    if rank is None:
        rank = g.op("Size", g.op("Shape", input))
    else:
        rank = g.op("Constant", value_t=torch.tensor(rank, dtype=torch.int64))
    extension = g.op(
        "Sub",
        g.op("Mul", rank, g.op("Constant", value_t=torch.tensor(2, dtype=torch.int64))),
        pad_len,
    )
    # Concat pad with extension: paddings = [dim_n_begin, dim_n_end,
    # dim_n-1_begin, dim_n-1_end, 0, 0, ... ]
    # Currently ONNX only supports int64 type for Pad
    pad = g.op("Cast", pad, to_i=sym_help.cast_pytorch_to_onnx["Long"])
    paddings = g.op(
        "Concat",
        pad,
        g.op(
            "ConstantOfShape", extension, value_t=torch.tensor([0], dtype=torch.int64)
        ),
        axis_i=0,
    )
    # Reshape and reverse order and collate first beginnings and then ends
    # paddings = [[..., 0, dim_n-1_begin, dim_n_begin],
    #               [..., 0, dim_n-1_end, dim_n_end]]
    # Reshape back to 1-D paddings = [..., 0, dim_n - 1_begin, dim_n_begin,
    # ..., 0, dim_n - 1_end, dim_n_end]

    # replace original Constant-Transpose-Constant with Slices and Concat.
    paddings = torch.onnx.symbolic_opset10.flip(g, paddings, [0])
    begins = sym_help._slice_helper(
        g, paddings, axes=[0], starts=[1], ends=[0xFFFF], steps=[2]
    )
    ends = sym_help._slice_helper(
        g, paddings, axes=[0], starts=[0], ends=[0xFFFF], steps=[2]
    )
    paddings = g.op("Concat", begins, ends, axis_i=0)
    padding_c = g.op("Cast", paddings, to_i=sym_help.cast_pytorch_to_onnx["Long"])
    return padding_c



# Copyright (c) OpenMMLab. All rights reserved.
import types
from collections import defaultdict
from typing import Any, Callable, Dict, List, MutableSequence, Optional, Tuple, Union

from juxtapose.mmdeploy.utils import IR, Backend, get_root_logger
from .rewriter_utils import (
    Checker,
    ContextCaller,
    RewriterRegistry,
    copy_function,
    get_frame_func,
    get_func_qualname,
    import_function,
)

try:
    try:
        # torch>=1.10.0
        from torch.fx._symbolic_trace import _wrapped_fns_to_patch
    except ImportError:
        # 1.10.0>torch>=1.8.0
        from torch.fx.symbolic_trace import _wrapped_fns_to_patch
except ImportError:
    # torch<1.8.0
    _wrapped_fns_to_patch = []


def _replace_all_obj(
    obj: Any,
    new_obj: Any,
    ignore_refs: Tuple[Any] = tuple(),
    ignore_keys: Tuple[str] = tuple(),
):
    """Replace all object reference with new_object.

    Args:
        obj (Any): The object to be replaced.
        new_obj (Any): The object to replace obj.
        ignore_refs (Tuple[Any]): These refs will be ignored.
        ignore_keys (Tuple[str]): object with these keys will be ignored.
    """
    import gc

    refs = gc.get_referrers(obj)
    obj_id = id(obj)
    for ref in refs:
        if ref in ignore_refs:
            continue
        elif isinstance(ref, MutableSequence):
            for i, v in enumerate(ref):
                if id(v) == obj_id:
                    ref[i] = new_obj
        elif isinstance(ref, Dict):
            for k, v in ref.items():
                if id(v) == obj_id and k not in ignore_keys:
                    ref[k] = new_obj
        else:
            # TODO: check if we can replace tuple
            pass


def _set_func(
    origin_func_path: str,
    rewrite_func: Callable,
    ignore_refs: Tuple[Any] = tuple(),
    ignore_keys: Tuple[str] = ("origin_func",),
):
    """Rewrite a function by executing a python statement.

    Args:
        origin_func_path (str): The path to origin function.
        rewrite_func (Callable): The new function instance.
        ignore_refs (Tuple[Any]): These refs will be ignored.
        ignore_keys (Tuple[str]): object with these keys will be ignored.
    """

    # Import necessary module
    split_path = origin_func_path.split(".")
    for i in range(len(split_path), 0, -1):
        try:
            exec("import {}".format(".".join(split_path[:i])))
            break
        except Exception:
            continue
    origin_func = eval(origin_func_path)
    method_class = False
    if len(split_path) > 1:
        module_or_class = eval(".".join(split_path[:-1]))
        if isinstance(module_or_class, type):
            method_class = True
    # Assign function
    if not method_class:
        _replace_all_obj(
            origin_func, rewrite_func, ignore_refs=ignore_refs, ignore_keys=ignore_keys
        )
    exec(f"{origin_func_path} = rewrite_func")


def _del_func(path: str):
    """Delete a function that is denoted by a path.

    Args:
        path (str): The path to evaluate.
    """

    split_path = path.split(".")
    for i in range(len(split_path), 0, -1):
        try:
            exec("import {}".format(".".join(split_path[:i])))
            exec(f"del {path}")
            break
        except Exception:
            continue


def _fx_wrap_copied_fn(func: types.FunctionType, copied_func: types.FunctionType):
    """If a function is wrapped by torch.fx.wrap, its copy also needs to be
    wrapped by torch.fx.wrap."""
    if not hasattr(func, "__globals__"):
        return

    wrapped_fns_globals = [item[0] for item in _wrapped_fns_to_patch]
    wrapped_fns_names = [item[1] for item in _wrapped_fns_to_patch]

    # check if wrapped by torch.fx.wrap
    if func.__globals__ in wrapped_fns_globals:
        idx = wrapped_fns_globals.index(func.__globals__)
        fn_name = wrapped_fns_names[idx]
        # a hacky way to wrap the func in copied func
        _wrapped_fns_to_patch.append((copied_func.__globals__, fn_name))


class FunctionRewriter:
    """A function rewriter which maintains rewritten functions.

    The rewritten functions can be registered by calling register_rewriter().
    In RewriteContext, the rewriter automatically replaces target functions and
    recovers them after exiting the context.

    Examples:
        >>> @FUNCTION_REWRITER.register_rewriter(
        >>>     func_name='torch.Tensor.size', backend='ncnn')
        >>> def size_of_tensor_static(self, *args):
        >>>     ctx = FUNCTION_REWRITER.get_context()
        >>>     ret = ctx.origin_func(self, *args)
        >>>     if isinstance(ret, torch.Tensor):
        >>>         ret = int(ret)
        >>>     else:
        >>>         ret = [int(r) for r in ret]
        >>>         ret = tuple(ret)
        >>>     return ret
    """

    def __init__(self):
        self._registry = RewriterRegistry()
        self._func_contexts = defaultdict(list)

    def register_rewriter(
        self,
        func_name: str,
        backend: str = Backend.DEFAULT.value,
        ir: IR = IR.DEFAULT,
        extra_checkers: Optional[Union[Checker, List[Checker]]] = None,
        **kwargs,
    ):
        """The interface of function rewriter decorator.

        Args:
            func_name (str): The function name/path to rewrite.
            backend (str): The rewriter will be activated on which backend.
            ir (IR): The rewriter will be activated on which IR.
            extra_checkers (Checker | List[Checker] | None): Other requirements
                defined by Checker.

        Returns:
            Callable: The process of registering function.
        """

        return self._registry.register_object(
            func_name, backend, ir, extra_checkers, **kwargs
        )

    def enter(self, cfg: Dict = dict(), env: Dict = dict(), **kwargs):
        """The implementation of function rewrite."""
        self._func_contexts.clear()
        # Get current records
        functions_records = self._registry.get_records(env)
        # Get current fx wrapped func nums
        self._ori_fx_wrap_num = len(_wrapped_fns_to_patch)

        self._origin_functions = list()
        self._additional_functions = list()
        new_functions = list()
        for function_path, record_dict in functions_records:
            # Check if the origin function exists
            try:
                origin_func, origin_class = import_function(function_path)
            except Exception:
                origin_func = None
                logger = get_root_logger()
                logger.warning(
                    f"Can not find {function_path}, function rewrite will "
                    "not be applied"
                )

            # Only rewrite functions that exist
            if origin_func is not None:
                is_addition_function = False
                if origin_class is not None:
                    function_name = function_path.split(".")[-1]
                    try:
                        origin_class.__getattribute__(origin_class, function_name)
                    except Exception:
                        # The function is a method and it is derived from base
                        # class.
                        is_addition_function = True

                if is_addition_function:
                    self._additional_functions.append(function_path)

                # Save origin function
                self._origin_functions.append(
                    dict(func_path=function_path, origin_func=origin_func)
                )

                # Create context_caller
                rewrite_function = record_dict["_object"]
                # The func before and after copy has different globals
                rewrite_function = copy_function(rewrite_function)
                extra_kwargs = kwargs.copy()
                extra_kwargs.update(record_dict)
                context_caller = ContextCaller(
                    rewrite_function, origin_func, cfg, **extra_kwargs
                )
                # If there is a function wrapped by torch.fx.wrap in
                # rewrite_function's globals, we need to wrap the same name
                # function in copied function's globals.
                _fx_wrap_copied_fn(record_dict["_object"], context_caller.func)

                qualname = get_func_qualname(rewrite_function)
                self._func_contexts[qualname].append(context_caller)
                self._func_contexts[function_path].append(context_caller)

                # Cache new the function to avoid homonymic bug
                new_functions.append(
                    dict(func_path=function_path, origin_func=rewrite_function)
                )

        for func_dict in new_functions:
            function_path = func_dict["func_path"]
            new_function = func_dict["origin_func"]
            # Rewrite functions
            _set_func(function_path, new_function)

    def exit(self):
        """Recover the function rewrite."""
        # Restore _wrapped_fns_to_patch
        cur_fx_wrap_num = len(_wrapped_fns_to_patch)
        for _ in range(cur_fx_wrap_num - self._ori_fx_wrap_num):
            _wrapped_fns_to_patch.pop(-1)

        for func_dict in self._origin_functions:
            func_path = func_dict["func_path"]
            func = func_dict["origin_func"]
            _set_func(func_path, func)
        for func_path in self._additional_functions:
            _del_func(func_path)

        self._func_contexts.clear()

    def get_context(self, key: Optional[str] = None) -> ContextCaller:
        """Get the context of rewriter.

        Args:
            key: key to the context.

        Returns:
            ContextCaller: context of function
        """
        func = None
        if key is None:
            func = get_frame_func(2)
            key = get_func_qualname(func)

        # get all contexts
        ctxs = self._func_contexts.get(key, [])

        if func is None:
            assert len(ctxs) == 1
            return ctxs[0]

        ctx = None
        for tmp_ctx in ctxs:
            if tmp_ctx.func == func:
                ctx = tmp_ctx

        if ctx is None:
            get_root_logger().warning(f"Can not found context of {key}")
        return ctx



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.expand", backend="ncnn")
def expand__ncnn(self, *sizes) -> torch.Tensor:
    """Rewrite `expand` for NCNN backend.

    Do not expand on batch dim for tensor with ndim >= 3
    """
    ctx = FUNCTION_REWRITER.get_context()
    if self.ndim < 3 or sizes[0] not in [1, -1]:
        return ctx.origin_func(*sizes)
    return self



# Copyright (c) OpenMMLab. All rights reserved.
from . import base_dense_head  # noqa: F401,F403
from . import centernet_head  # noqa: F401,F403
from . import detr_head  # noqa: F401,F403
from . import fovea_head  # noqa: F401,F403
from . import gfl_head  # noqa: F401,F403
from . import reppoints_head  # noqa: F401,F403
from . import rpn_head  # noqa: F401,F403
from . import rtmdet_head  # noqa: F401,F403
from . import rtmdet_ins_head  # noqa: F401,F403
from . import solo_head  # noqa: F401,F403
from . import solov2_head  # noqa: F401,F403
from . import yolo_head  # noqa: F401,F403
from . import yolox_head  # noqa: F401,F403



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Sequence

import torch


class BaseWrapper(torch.nn.Module, metaclass=ABCMeta):
    """Abstract base class for backend wrappers.

    Args:
        output_names (Sequence[str]): Names to model outputs in order, which is
        useful when converting the output dict to a ordered list or converting
        the output ordered list to a key-value dict.
    """

    def __init__(self, output_names: Sequence[str]):
        super().__init__()
        self._output_names = output_names

    @staticmethod
    def get_backend_file_count() -> int:
        """Return the count of backend file(s)

        Each backend has its own requirement on backend files (e.g., TensorRT
        requires 1 .engine file and ncnn requires 2 files (.param, .bin)). This
        interface allow developers to get the count of these required files.

        Returns:
            int: The count of required backend file(s).
        """
        return 1

    @abstractmethod
    def forward(self, inputs: Dict[str,
                                   torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Run forward inference.

        Args:
            inputs (Dict[str, torch.Tensor]): Key-value pairs of model inputs.

        Returns:
            Dict[str, torch.Tensor]: Key-value pairs of model outputs.
        """
        pass

    @property
    def output_names(self):
        """Return the output names."""
        return self._output_names

    @output_names.setter
    def output_names(self, value):
        """Set the output names."""
        self._output_names = value

    def output_to_list(self, output_dict: Dict[str,
                                               torch.Tensor]) -> \
            List[torch.Tensor]:
        """Convert the output dict of forward() to a tensor list.

        Args:
            output_dict (Dict[str, torch.Tensor]): Key-value pairs of model
                outputs.

        Returns:
            List[torch.Tensor]: An output value list whose order is determined
                by the ouput_names list.
        """
        outputs = [output_dict[name] for name in self._output_names]
        return outputs



# Copyright (c) OpenMMLab. All rights reserved.
from .deploy import (ObjectDetection, clip_bboxes, gather_topk,
                     get_post_processing_params, pad_with_value,
                     pad_with_value_if_necessary)

__all__ = [
    'get_post_processing_params', 'clip_bboxes', 'pad_with_value',
    'pad_with_value_if_necessary', 'ObjectDetection', 'gather_topk'
]



# Copyright (c) OpenMMLab. All rights reserved.
from torch.utils.data import Dataset


def is_can_sort_dataset(dataset: Dataset) -> bool:
    """Checking for the possibility of sorting the dataset by fields 'height'
    and 'width'.

    Args:
        dataset (Dataset): The dataset.

    Returns:
        bool: Is it possible or not to sort the dataset.
    """
    is_sort_possible = \
        hasattr(dataset, 'data_infos') and \
        dataset.data_infos and \
        all(key in dataset.data_infos[0] for key in ('height', 'width'))
    return is_sort_possible


def sort_dataset(dataset: Dataset) -> Dataset:
    """Sorts the dataset by image height and width.

    Args:
        dataset (Dataset): The dataset.

    Returns:
        Dataset: Sorted dataset.
    """
    sort_data_infos = sorted(
        dataset.data_infos, key=lambda e: (e['height'], e['width']))
    sort_img_ids = [e['id'] for e in sort_data_infos]
    dataset.data_infos = sort_data_infos
    dataset.img_ids = sort_img_ids
    return dataset



# Copyright (c) OpenMMLab. All rights reserved.
from torch.onnx import symbolic_helper

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER
from juxtapose.mmdeploy.utils import Backend


@symbolic_helper.parse_args("v")
def gelu__ncnn_pt111(g, self):
    """gelu for torch<=1.12."""
    return g.op("mmdeploy::Gelu", self)


@SYMBOLIC_REWRITER.register_symbolic(
    "gelu", is_pytorch=True, backend=Backend.NCNN.value
)
def gelu__ncnn(g, self, approximate: str = "none"):
    """Support export GELU with ncnn backend."""
    return gelu__ncnn_pt111(g, self)



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.utils import IR


@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.chunk", backend="ncnn")
def chunk__ncnn(self, num_chunks: int, dim: int = 0) -> torch.Tensor:
    """Rewrite `chunk` for NCNN backend.

    Chunk in ncnn are not supported, so it should be rewritten.
    """
    dim_len = self.shape[dim]
    # int ceil.
    step = dim_len // num_chunks
    if dim_len % num_chunks > 0:
        step += 1
    index_list = []
    index = 0
    while index < dim_len:
        index_list.append(index)
        index += step
    index_list.append(dim_len)
    output = [
        self.index_select(
            dim,
            torch.tensor(
                [j for j in range(index_list[i], index_list[i + 1])], dtype=torch.int64
            ),
        )
        for i in range(len(index_list) - 1)
    ]

    return output


@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.chunk", ir=IR.TORCHSCRIPT)
def chunk__torchscript(self, num_chunks: int, dim: int = 0) -> torch.Tensor:
    """Rewrite `chunk` for Torchscript.

    Replace chunk op with split op
    """
    dim_size = self.shape[dim]
    assert dim_size % num_chunks == 0, "cannot split to equal sizes"
    output = self.split(dim_size // num_chunks, dim=dim)
    return output



# Copyright (c) OpenMMLab. All rights reserved.
import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.size", backend="ncnn")
def tensor__size__ncnn(self, *args):
    """Rewrite `size` for ncnn backend.

    ONNX Shape node is not supported in ncnn. This function return integer
    instead of Torch.Size to avoid ONNX Shape node.
    """
    ctx = FUNCTION_REWRITER.get_context()

    ret = ctx.origin_func(self, *args)
    if isinstance(ret, torch.Tensor):
        ret = int(ret)
    elif isinstance(ret, int):
        return ret
    else:
        ret = [int(r) for r in ret]
        ret = tuple(ret)
    return ret


@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.size", backend="ascend")
def tensor__size__ascend(self, *args):
    """Rewrite `size` for ascens backend.

    Support negative index.
    """
    ctx = FUNCTION_REWRITER.get_context()

    if len(args) != 0:
        index = args[0]
        if index < 0:
            index = self.dim() + index
            args = (index,)

    return ctx.origin_func(self, *args)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence

import numpy as np
import torch
from mmdet.utils import OptConfigType
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet.deploy import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.mmcv.ops import multiclass_nms
from juxtapose.mmdeploy.utils import Backend, is_dynamic_shape


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.yolo_head." "YOLOV3Head.predict_by_feat"
)
def yolov3_head__predict_by_feat(
    self,
    pred_maps: Sequence[Tensor],
    cfg: OptConfigType = None,
    rescale: bool = False,
    with_nms: bool = True,
    **kwargs,
):
    """Rewrite `predict_by_feat` of `YOLOV3Head` for default backend.

    Rewrite this function to deploy model, transform network output for a
    batch into bbox predictions.

    Args:
        ctx (ContextCaller): The context with additional information.
        pred_maps (Sequence[Tensor]): Raw predictions for a batch of
            images.
        cfg (ConfigDict, optional): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.

    Returns:
        If with_nms == True:
            tuple[Tensor, Tensor]: tuple[Tensor, Tensor]: (dets, labels),
            `dets` of shape [N, num_det, 5] and `labels` of shape
            [N, num_det].
        Else:
            tuple[Tensor, Tensor]: batch_mlvl_bboxes, batch_mlvl_scores
    """
    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg

    # mark pred_maps
    @mark("yolo_head", inputs=["pred_maps"])
    def __mark_pred_maps(pred_maps):
        return pred_maps

    pred_maps = __mark_pred_maps(pred_maps)

    is_dynamic_flag = is_dynamic_shape(ctx.cfg)
    num_levels = len(pred_maps)
    pred_maps_list = [pred_maps[i].detach() for i in range(num_levels)]

    cfg = self.test_cfg if cfg is None else cfg
    assert len(pred_maps_list) == self.num_levels

    device = pred_maps_list[0].device
    batch_size = pred_maps_list[0].shape[0]

    featmap_sizes = [pred_maps_list[i].shape[-2:] for i in range(self.num_levels)]
    multi_lvl_anchors = self.prior_generator.grid_anchors(featmap_sizes, device)
    pre_topk = cfg.get("nms_pre", -1)
    multi_lvl_bboxes = []
    multi_lvl_cls_scores = []
    multi_lvl_conf_scores = []
    for i in range(self.num_levels):
        # get some key info for current scale
        pred_map = pred_maps_list[i]
        stride = self.featmap_strides[i]
        # (b,h, w, num_anchors*num_attrib) ->
        # (b,h*w*num_anchors, num_attrib)
        pred_map = pred_map.permute(0, 2, 3, 1).reshape(batch_size, -1, self.num_attrib)
        # Inplace operation like
        # ```pred_map[..., :2] = \torch.sigmoid(pred_map[..., :2])```
        # would create constant tensor when exporting to onnx
        pred_map_conf = torch.sigmoid(pred_map[..., :2])
        pred_map_rest = pred_map[..., 2:]
        pred_map = torch.cat([pred_map_conf, pred_map_rest], dim=-1)
        pred_map_boxes = pred_map[..., :4]
        multi_lvl_anchor = multi_lvl_anchors[i]
        # use static anchor if input shape is static
        if not is_dynamic_flag:
            multi_lvl_anchor = multi_lvl_anchor.data
        multi_lvl_anchor = multi_lvl_anchor.unsqueeze(0)
        bbox_pred = self.bbox_coder.decode(multi_lvl_anchor, pred_map_boxes, stride)
        # conf and cls
        conf_pred = torch.sigmoid(pred_map[..., 4])
        cls_pred = torch.sigmoid(pred_map[..., 5:]).view(
            batch_size, -1, self.num_classes
        )  # Cls pred one-hot.

        # Save the result of current scale
        multi_lvl_bboxes.append(bbox_pred)
        multi_lvl_cls_scores.append(cls_pred)
        multi_lvl_conf_scores.append(conf_pred)
    # Merge the results of different scales together
    batch_mlvl_bboxes = torch.cat(multi_lvl_bboxes, dim=1)
    batch_mlvl_scores = torch.cat(multi_lvl_cls_scores, dim=1)
    batch_mlvl_conf_scores = torch.cat(multi_lvl_conf_scores, dim=1)

    post_params = get_post_processing_params(deploy_cfg)

    if pre_topk > 0:
        _, topk_inds = conf_pred.topk(pre_topk)
        batch_inds = torch.arange(batch_size, device=device).unsqueeze(-1).long()
        # Avoid onnx2tensorrt issue in https://github.com/NVIDIA/TensorRT/issues/1134 # noqa: E501
        transformed_inds = bbox_pred.shape[1] * batch_inds + topk_inds.long()
        bbox_pred = bbox_pred.reshape(-1, 4)[transformed_inds, :].reshape(
            batch_size, -1, 4
        )
        cls_pred = cls_pred.reshape(-1, self.num_classes)[transformed_inds, :].reshape(
            batch_size, -1, self.num_classes
        )
        conf_pred = conf_pred.reshape(-1, 1)[transformed_inds].reshape(batch_size, -1)

    batch_mlvl_conf_scores = batch_mlvl_conf_scores.unsqueeze(2)
    batch_mlvl_scores = batch_mlvl_scores * batch_mlvl_conf_scores
    if with_nms:
        max_output_boxes_per_class = post_params.max_output_boxes_per_class
        iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
        pre_top_k = post_params.pre_top_k
        keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
        # keep aligned with original pipeline, improve
        # mAP by 1% for YOLOv3 in ONNX
        score_threshold = 0
        nms_type = cfg.nms.get("type")
        return multiclass_nms(
            batch_mlvl_bboxes,
            batch_mlvl_scores,
            max_output_boxes_per_class,
            nms_type=nms_type,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            pre_top_k=pre_top_k,
            keep_top_k=keep_top_k,
        )
    else:
        return batch_mlvl_bboxes, batch_mlvl_scores


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdet.models.dense_heads.YOLOV3Head.predict_by_feat",
    backend=Backend.NCNN.value,
)
def yolov3_head__predict_by_feat__ncnn(
    self, pred_maps, with_nms=True, cfg=None, **kwargs
):
    """Rewrite `predict_by_feat` of YOLOV3Head for ncnn backend.

    1. Shape node and batch inference is not supported by ncnn. This function
    transform dynamic shape to constant shape and remove batch inference.
    2. Batch dimension is not supported by ncnn, but supported by pytorch.
    The negative value of axis in torch.cat is rewritten as corresponding
    positive value to avoid axis shift.
    3. 2-dimension tensor broadcast of `BinaryOps` operator is not supported by
    ncnn. This function unsqueeze 2-dimension tensor to 3-dimension tensor for
    correct `BinaryOps` calculation by ncnn.


    Args:
        ctx (ContextCaller): The context with additional information.

        self: Represent the instance of the original class.
        pred_maps (list[Tensor]): Raw predictions for a batch of images.
        with_nms (bool): If True, do nms before return boxes.
            Default: True.
        cfg (mmengine.Config | None): Test / postprocessing configuration,
            if None, test_cfg would be used. Default: None.

    Returns:
            Tensor: Detection_output of shape [num_boxes, 6],
            each row is [label, score, x1, y1, x2, y2]. Note that
            fore-ground class label in Yolov3DetectionOutput starts
            from `1`. x1, y1, x2, y2 are normalized in range(0,1).
    """
    ctx = FUNCTION_REWRITER.get_context()
    num_levels = len(pred_maps)
    cfg = self.test_cfg if cfg is None else cfg
    post_params = get_post_processing_params(ctx.cfg)

    confidence_threshold = cfg.get("conf_thr", post_params.confidence_threshold)
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    anchor_biases = np.array(self.prior_generator.base_sizes).reshape(-1).tolist()
    num_box = len(self.prior_generator.base_sizes[0])
    bias_masks = list(range(num_levels * num_box))

    def _create_yolov3_detection_output():
        """Help create Yolov3DetectionOutput op in ONNX."""

        class Yolov3DetectionOutputOp(torch.autograd.Function):
            """Create Yolov3DetectionOutput op.

            Args:
                *inputs (Tensor): Multiple predicted feature maps.
                num_class (int): Number of classes.
                num_box (int): Number of box per grid.
                confidence_threshold (float): Threshold of object
                    score.
                nms_threshold (float): IoU threshold for NMS.
                biases (List[float]: Base sizes to compute anchors
                    for each FPN.
                mask (List[float]): Used to select base sizes in
                    biases.
                anchors_scale (List[float]): Down-sampling scales of
                    each FPN layer, e.g.: [32, 16].
            """

            @staticmethod
            def forward(ctx, *args):
                # create dummpy output of shape [num_boxes, 6],
                # each row is [label, score, x1, y1, x2, y2]
                output = torch.rand(100, 6)
                return output

            @staticmethod
            def symbolic(g, *args):
                anchors_scale = args[-1]
                inputs = args[: len(anchors_scale)]
                assert len(args) == (len(anchors_scale) + 7)
                return g.op(
                    "mmdeploy::Yolov3DetectionOutput",
                    *inputs,
                    num_class_i=args[-7],
                    num_box_i=args[-6],
                    confidence_threshold_f=args[-5],
                    nms_threshold_f=args[-4],
                    biases_f=args[-3],
                    mask_f=args[-2],
                    anchors_scale_f=anchors_scale,
                    outputs=1,
                )

        return Yolov3DetectionOutputOp.apply(
            *pred_maps,
            self.num_classes,
            num_box,
            confidence_threshold,
            iou_threshold,
            anchor_biases,
            bias_masks,
            self.featmap_strides,
        )

    output = _create_yolov3_detection_output()
    return output
